"""Library for GEO values."""

import pytest
from ical.exceptions import CalendarParseError

from ical.component import ComponentModel
from ical.parsing.property import ParsedProperty
from ical.types.geo import Geo


def test_geo() -> None:
    """Test for geo fields."""

    class TestModel(ComponentModel):
        """Model under test."""

        geo: Geo

    model = TestModel.parse_obj(
        {"geo": [ParsedProperty(name="geo", value="120.0;-30.1")]}
    )
    assert model.geo.lat == 120.0
    assert model.geo.lng == -30.1

    with pytest.raises(CalendarParseError):
        TestModel.parse_obj({"geo": [ParsedProperty(name="geo", value="10")]})



"""Library for managing the lifecycle of components in a calendar.

A store is like a manager for events within a Calendar, updating the necessary
properties such as modification times, sequence numbers, and ids. This higher
level API is a more convenient API than working with the lower level objects
directly.
"""

# pylint: disable=unnecessary-lambda

from __future__ import annotations

import datetime
import logging
from collections.abc import Callable, Iterable, Generator
from typing import Any, TypeVar, Generic, cast

from .calendar import Calendar
from .event import Event
from .exceptions import StoreError, TodoStoreError, EventStoreError
from .iter import RulesetIterable
from .list import todo_list_view
from .timezone import Timezone
from .todo import Todo
from .types import Range, Recur, RecurrenceId, RelationshipType
from .tzif.timezoneinfo import TimezoneInfoError
from .util import dtstamp_factory, local_timezone


_LOGGER = logging.getLogger(__name__)


__all__ = [
    "EventStore",
    "EventStoreError",
    "TodoStore",
    "TodoStoreError",
    "StoreError",
]

_T = TypeVar("_T", bound="Event | Todo")


def _ensure_timezone(
    dtstart: datetime.datetime | datetime.date | None, timezones: list[Timezone]
) -> Timezone | None:
    """Create a timezone object for the specified date if it does not already exist."""
    if (
        not isinstance(dtstart, datetime.datetime)
        or not dtstart.utcoffset()
        or not dtstart.tzinfo
    ):
        return None

    # Verify this timezone does not already exist. The number of timezones
    # in a calendar is typically very small so iterate over the whole thing
    # to avoid any synchronization/cache issues.
    key = str(dtstart.tzinfo)
    for timezone in timezones:
        if timezone.tz_id == key:
            return None

    try:
        return Timezone.from_tzif(key)
    except TimezoneInfoError as err:
        raise EventStoreError(
            f"No timezone information available for event: {key}"
        ) from err


def _match_item(item: _T, uid: str, recurrence_id: str | None) -> bool:
    """Return True if the item is an instance of a recurring event."""
    if item.uid != uid:
        return False
    if recurrence_id is None:
        # Match all items with the specified uids
        return True
    # Match a single item with the specified recurrence_id. If the item is an
    # edited instance match return it
    if item.recurrence_id == recurrence_id:
        _LOGGER.debug("Matched exact recurrence_id: %s", item)
        return True
    # Otherwise, determine if this instance is in the series
    _LOGGER.debug(
        "Expanding item %s %s to look for match of %s", uid, item.dtstart, recurrence_id
    )
    dtstart = RecurrenceId.to_value(recurrence_id)
    for dt in item.as_rrule() or ():
        if dt == dtstart:
            _LOGGER.debug("Found expanded recurrence_id: %s", dt)
            return True
    return False


def _match_items(
    items: list[_T], uid: str, recurrence_id: str | None
) -> Generator[tuple[int, _T], None, None]:
    """Return items from the list that match the uid and recurrence_id."""
    for index, item in enumerate(items):
        if _match_item(item, uid, recurrence_id):
            yield index, item


def _prepare_update(
    store_item: Event | Todo,
    item: Event | Todo,
    recurrence_id: str | None = None,
    recurrence_range: Range = Range.NONE,
) -> dict[str, Any]:
    """Prepare an update to an existing event."""
    partial_update = item.dict(
        exclude_unset=True,
        exclude={"dtstamp", "uid", "sequence", "created", "last_modified"},
    )
    _LOGGER.debug("Preparing update update=%s", item)
    update = {
        "created": store_item.dtstamp,
        "sequence": (store_item.sequence + 1) if store_item.sequence else 1,
        "last_modified": item.dtstamp,
        **partial_update,
        "dtstamp": item.dtstamp,
    }
    if rrule := update.get("rrule"):
        update["rrule"] = Recur.parse_obj(rrule)
    if recurrence_id and store_item.rrule:
        # Forking a new event off the old event preserves the original uid and
        # recurrence_id.
        update.update(
            {
                "uid": store_item.uid,
                "recurrence_id": recurrence_id,
            }
        )
        if recurrence_range == Range.NONE:
            # The new event copied from the original is a single instance,
            # which is not recurring.
            update["rrule"] = None
        else:
            # Overwriting with a new recurring event
            update["created"] = item.dtstamp

            # Adjust start and end time of the event
            dtstart: datetime.datetime | datetime.date = RecurrenceId.to_value(
                recurrence_id
            )
            if item.dtstart:
                dtstart = item.dtstart
            update["dtstart"] = dtstart
            # Event either has a duration (which should already be set) or has
            # an explicit end which needs to be realigned to new start time.
            if isinstance(store_item, Event) and store_item.dtend:
                update["dtend"] = dtstart + store_item.computed_duration
    return update


class GenericStore(Generic[_T]):
    """A a store manages the lifecycle of items on a Calendar."""

    def __init__(
        self,
        items: list[_T],
        timezones: list[Timezone],
        exc: type[StoreError],
        dtstamp_fn: Callable[[], datetime.datetime] = lambda: dtstamp_factory(),
        tzinfo: datetime.tzinfo | None = None,
    ):
        """Initialize the EventStore."""
        self._items = items
        self._timezones = timezones
        self._exc = exc
        self._dtstamp_fn = dtstamp_fn
        self._tzinfo = tzinfo or local_timezone()

    def add(self, item: _T) -> _T:
        """Add the specified item to the calendar.

        This will handle assigning modification dates, sequence numbers, etc
        if those fields are unset.

        The store will ensure the `ical.calendar.Calendar` has the necessary
        `ical.timezone.Timezone` needed to fully specify the time information
        when encoded.
        """
        update: dict[str, Any] = {}
        if not item.created:
            update["created"] = item.dtstamp
        if item.sequence is None:
            update["sequence"] = 0
        if isinstance(item, Todo) and not item.dtstart:
            if item.due:
                update["dtstart"] = item.due - datetime.timedelta(days=1)
            else:
                update["dtstart"] = datetime.datetime.now(tz=self._tzinfo)
        new_item = cast(_T, item.copy_and_validate(update=update))

        # The store can only manage cascading deletes for some relationship types
        for relation in new_item.related_to or ():
            if relation.reltype != RelationshipType.PARENT:
                raise self._exc(f"Unsupported relationship type {relation.reltype}")

        _LOGGER.debug("Adding item: %s", new_item)
        self._ensure_timezone(item.dtstart)
        self._items.append(new_item)
        return new_item

    def delete(
        self,
        uid: str,
        recurrence_id: str | None = None,
        recurrence_range: Range = Range.NONE,
    ) -> None:
        """Delete the item from the calendar.

        This method is used to delete an existing item. For a recurring item
        either the whole item or instances of an item may be deleted. To
        delete the complete range of a recurring item, the `uid` property
        for the item must be specified and the `recurrence_id` should not
        be specified. To delete an individual instances of the item the
        `recurrence_id` must be specified.

        When deleting individual instances, the range property may specify
        if deletion of just a specific instance, or a range of instances.
        """
        items_to_delete: list[_T] = [
            item for _, item in _match_items(self._items, uid, recurrence_id)
        ]
        if not items_to_delete:
            raise self._exc(
                f"No existing item with uid/recurrence_id: {uid}/{recurrence_id}"
            )

        for store_item in items_to_delete:
            self._apply_delete(store_item, recurrence_id, recurrence_range)

    def _apply_delete(
        self,
        store_item: _T,
        recurrence_id: str | None = None,
        recurrence_range: Range = Range.NONE,
    ) -> None:
        if (
            recurrence_id
            and recurrence_range == Range.THIS_AND_FUTURE
            and RecurrenceId.to_value(recurrence_id) == store_item.dtstart
        ):
            # Editing the first instance and all forward is the same as editing the
            # entire series so don't bother forking a new event
            recurrence_id = None

        children = []
        for event in self._items:
            for relation in event.related_to or ():
                if (
                    relation.reltype == RelationshipType.PARENT
                    and relation.uid == store_item.uid
                ):
                    children.append(event)
        for child in children:
            self._items.remove(child)

        # Deleting all instances in the series
        if not recurrence_id or not store_item.rrule:
            self._items.remove(store_item)
            return

        exdate = RecurrenceId.to_value(recurrence_id)
        if recurrence_range == Range.NONE:
            # A single recurrence instance is removed. Add an exclusion to
            # to the event.
            store_item.exdate.append(exdate)
            return

        # Assumes any recurrence deletion is valid, and that overwriting
        # the "until" value will not produce more instances. UNTIL is
        # inclusive so it can't include the specified exdate. FREQ=DAILY
        # is the lowest frequency supported so subtracting one day is
        # safe and works for both dates and datetimes.
        store_item.rrule.count = None
        store_item.rrule.until = exdate - datetime.timedelta(days=1)
        now = self._dtstamp_fn()
        store_item.dtstamp = now
        store_item.last_modified = now

    def edit(
        self,
        uid: str,
        item: _T,
        recurrence_id: str | None = None,
        recurrence_range: Range = Range.NONE,
    ) -> None:
        """Update the item with the specified uid.

        The specified item should be created with minimal fields, just
        including the fields that should be updated. The default fields such
        as `uid` and `dtstamp` may be used to set the uid for a new created item
        when updating a recurring item, or for any modification times.

        For a recurring item, either the whole item or individual instances
        of the item may be edited. To edit the complete range of a recurring
        item the `uid` property must be specified and the `recurrence_id` should
        not be specified. To edit an individual instances of the item the
        `recurrence_id` must be specified. The `recurrence_range` determines if
        just that individual instance is updated or all items following as well.

        The store will ensure the `ical.calendar.Calendar` has the necessary
        `ical.timezone.Timezone` needed to fully specify the item time information
        when encoded.
        """
        items_to_edit: list[tuple[int, _T]] = [
            (index, item)
            for index, item in _match_items(self._items, uid, recurrence_id)
        ]
        if not items_to_edit:
            raise self._exc(
                f"No existing item with uid/recurrence_id: {uid}/{recurrence_id}"
            )

        for store_index, store_item in items_to_edit:
            self._apply_edit(
                store_index, store_item, item, recurrence_id, recurrence_range
            )

    def _apply_edit(
        self,
        store_index: int,
        store_item: _T,
        item: _T,
        recurrence_id: str | None = None,
        recurrence_range: Range = Range.NONE,
    ) -> None:
        if (
            recurrence_id
            and recurrence_range == Range.THIS_AND_FUTURE
            and RecurrenceId.to_value(recurrence_id) == store_item.dtstart
        ):
            # Editing the first instance and all forward is the same as editing the
            # entire series so don't bother forking a new item
            recurrence_id = None

        update = _prepare_update(store_item, item, recurrence_id, recurrence_range)
        if recurrence_range == Range.NONE:
            # Changing the recurrence rule of a single item in the middle of the series
            # is not allowed. It is allowed to convert a single instance item to recurring.
            if item.rrule and store_item.rrule:
                if item.rrule.as_rrule_str() != store_item.rrule.as_rrule_str():
                    raise self._exc(
                        f"Can't update single instance with rrule (rrule={item.rrule})"
                    )
                item.rrule = None

        # Make a deep copy since deletion may update this objects recurrence rules
        new_item = cast(_T, store_item.copy_and_validate(update=update))
        if (
            recurrence_id
            and new_item.rrule
            and new_item.rrule.count
            and store_item.dtstart
        ):
            # The recurring item count needs to skip any items that
            # come before the start of the new item. Use a RulesetIterable
            # to handle workarounds for dateutil.rrule limitations.
            dtstart: datetime.date | datetime.datetime = update["dtstart"]
            ruleset = RulesetIterable(
                store_item.dtstart,
                [new_item.rrule.as_rrule(store_item.dtstart)],
                [],
                [],
            )
            for dtvalue in ruleset:
                if dtvalue >= dtstart:
                    break
                new_item.rrule.count = new_item.rrule.count - 1

        # The store can only manage cascading deletes for some relationship types
        for relation in new_item.related_to or ():
            if relation.reltype != RelationshipType.PARENT:
                raise self._exc(f"Unsupported relationship type {relation.reltype}")

        self._ensure_timezone(new_item.dtstart)

        # Editing a single instance of a recurring item is like deleting that instance
        # then adding a new instance on the specified date. If recurrence id is not
        # specified then the entire item is replaced.
        self.delete(
            store_item.uid,
            recurrence_id=recurrence_id,
            recurrence_range=recurrence_range,
        )
        self._items.insert(store_index, new_item)

    def _ensure_timezone(
        self, dtstart: datetime.datetime | datetime.date | None
    ) -> None:
        if (new_timezone := _ensure_timezone(dtstart, self._timezones)) is not None:
            self._timezones.append(new_timezone)


class EventStore(GenericStore[Event]):
    """An event store manages the lifecycle of events on a Calendar.

    An `ical.calendar.Calendar` is a lower level object that can be directly
    manipulated to add/remove an `ical.event.Event`. That is, it does not
    handle updating timestamps, incrementing sequence numbers, or managing
    lifecycle of a recurring event during an update.


    Here is an example for setting up an `EventStore`:

    ```python
    import datetime
    from ical.calendar import Calendar
    from ical.event import Event
    from ical.store import EventStore
    from ical.types import Recur

    calendar = Calendar()
    store = EventStore(calendar)

    event = Event(
        summary="Event summary",
        start="2022-07-03",
        end="2022-07-04",
        rrule=Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
    )
    store.add(event)
    ```

    This will add events to the calendar:
    ```python3
    for event in calendar.timeline:
        print(event.summary, event.uid, event.recurrence_id, event.dtstart)
    ```
    With output like this:
    ```
    Event summary a521cf45-2c02-11ed-9e5c-066a07ffbaf5 20220703 2022-07-03
    Event summary a521cf45-2c02-11ed-9e5c-066a07ffbaf5 20220710 2022-07-10
    Event summary a521cf45-2c02-11ed-9e5c-066a07ffbaf5 20220717 2022-07-17
    ```

    You may also delete an event, or a specific instance of a recurring event:
    ```python
    # Delete a single instance of the recurring event
    store.delete(uid=event.uid, recurrence_id="20220710")
    ```

    Then viewing the store using the `print` example removes the individual
    instance in the event:
    ```
    Event summary a521cf45-2c02-11ed-9e5c-066a07ffbaf5 20220703 2022-07-03
    Event summary a521cf45-2c02-11ed-9e5c-066a07ffbaf5 20220717 2022-07-17
    ```

    Editing an event is also supported:
    ```python
    store.edit("event-uid-1", Event(summary="New Summary"))
    ```
    """

    def __init__(
        self,
        calendar: Calendar,
        dtstamp_fn: Callable[[], datetime.datetime] = lambda: dtstamp_factory(),
    ):
        """Initialize the EventStore."""
        super().__init__(
            calendar.events,
            calendar.timezones,
            EventStoreError,
            dtstamp_fn,
            tzinfo=None,
        )


class TodoStore(GenericStore[Todo]):
    """A To-do store manages the lifecycle of to-dos on a Calendar."""

    def __init__(
        self,
        calendar: Calendar,
        tzinfo: datetime.tzinfo | None = None,
        dtstamp_fn: Callable[[], datetime.datetime] = lambda: dtstamp_factory(),
    ):
        """Initialize the TodoStore."""
        super().__init__(
            calendar.todos,
            calendar.timezones,
            TodoStoreError,
            dtstamp_fn,
            tzinfo=tzinfo,
        )
        self._calendar = calendar

    def todo_list(self, dtstart: datetime.datetime | None = None) -> Iterable[Todo]:
        """Return a list of all todos on the calendar.

        This view accounts for recurring todos.
        """
        return todo_list_view(self._calendar.todos, dtstart)



"""Library for parsing and encoding rfc5545 components with pydantic.

The data model returned by the contentlines parsing is a bag of ParsedProperty
objects that support all the flexibility of the rfc5545 spec. However in the
common case the spec has a lot more flexibility than is needed for handling
simple property types e.g. a single summary field that is specified only once.

This library helps reduce boilerplate for translating that complex structure
into the simpler pydantic data model, and handles custom field types and
validators.

Just as the pydantic model provides syntax glue for parsing data and
associating necessary validators, this is also doing the same thing
in the opposite direction to encode back to ICS.
"""

from __future__ import annotations

import copy
import datetime
import json
import logging
from typing import Any, Union, get_args, get_origin

try:
    from pydantic.v1 import BaseModel, root_validator, ValidationError
    from pydantic.v1.fields import SHAPE_LIST
except ImportError:
    from pydantic import BaseModel, root_validator, ValidationError  # type: ignore[no-redef, assignment]
    from pydantic.fields import SHAPE_LIST  # type: ignore[attr-defined, no-redef]

from .parsing.component import ParsedComponent
from .parsing.property import ParsedProperty
from .types.data_types import DATA_TYPE
from .exceptions import CalendarParseError

_LOGGER = logging.getLogger(__name__)


ATTR_VALUE = "VALUE"

# Repeated values can either be specified as multiple separate values, but
# also some values support repeated values within a single value with a
# comma delimiter, listed here.
EXPAND_REPEATED_VALUES = {
    "categories",
    "classification",
    "exdate",
    "rdate",
    "resources",
    "freebusy",
}


def _adjust_recurrence_date(
    date_value: datetime.datetime | datetime.date,
    dtstart: datetime.datetime | datetime.date,
) -> datetime.datetime | datetime.date:
    """Apply fixes to the recurrence rule date."""
    if isinstance(dtstart, datetime.datetime):
        if not isinstance(date_value, datetime.datetime):
            raise ValueError(
                "DTSTART was DATE-TIME but UNTIL was DATE: "
                "must be the same value type"
            )
        if dtstart.tzinfo is None:
            if date_value.tzinfo is not None:
                raise ValueError("DTSTART is date local but UNTIL was not")
            return date_value

        if date_value.utcoffset():
            raise ValueError("DTSTART had UTC or local and UNTIL must be UTC")

        return date_value

    if isinstance(date_value, datetime.datetime):
        # Fix invalid rules where UNTIL value is DATE-TIME but DTSTART is DATE
        return date_value.date()

    return date_value


def validate_until_dtstart(_cls: BaseModel, values: dict[str, Any]) -> dict[str, Any]:
    """Verify the until time and dtstart are the same."""
    if (
        not (rule := values.get("rrule"))
        or not rule.until
        or not (dtstart := values.get("dtstart"))
    ):
        return values
    rule.until = _adjust_recurrence_date(rule.until, dtstart)
    return values


def _as_datetime(
    date_value: datetime.datetime | datetime.date,
    dtstart: datetime.datetime,
) -> datetime.datetime:
    if not isinstance(date_value, datetime.datetime):
        return datetime.datetime.combine(date_value, dtstart.time())
    return date_value

def _as_date(
    date_value: datetime.datetime | datetime.date,
    dtstart: datetime.datetime,
) -> datetime.date:
    if isinstance(date_value, datetime.datetime):
        return datetime.date.fromordinal(date_value.toordinal())
    return date_value


def validate_recurrence_dates(
    _cls: BaseModel, values: dict[str, Any]
) -> dict[str, Any]:
    """Verify the recurrence dates have the correct types."""
    if (
        not values.get("rrule")
        or not (dtstart := values.get("dtstart"))
        or not (isinstance(dtstart, datetime.datetime) or isinstance(dtstart, datetime.date))
    ):
        return values
    is_datetime = isinstance(dtstart, datetime.datetime)
    validator = _as_datetime if is_datetime else _as_date
    for field in ("exdate", "rdate"):
        if not (date_values := values.get(field)):
            continue

        values[field] = [
            validator(date_value, dtstart) for date_value in date_values
        ]
    return values


class ComponentModel(BaseModel):
    """Abstract class for rfc5545 component model."""

    def __init__(self, **data: Any) -> None:
        try:
            super().__init__(**data)
        except ValidationError as err:
            raise CalendarParseError(f"Failed to parse component: {err}") from err

    def copy_and_validate(self, update: dict[str, Any]) -> ComponentModel:
        """Create a new object with updated values and validate it."""
        # Make a deep copy since deletion may update this objects recurrence rules
        new_item_copy = self.copy(update=update, deep=True)
        # Create a new object using the constructore to ensure we're performing
        # validation on the new object.
        return self.__class__(**new_item_copy.dict())

    @root_validator(pre=True, allow_reuse=True)
    def parse_extra_fields(
        cls, values: dict[str, list[ParsedProperty | ParsedComponent]]
    ) -> dict[str, Any]:
        """Parse extra fields not in the model."""
        all_fields = set()
        for field in cls.__fields__.values():
            all_fields |= {field.alias, field.name}

        extras: list[ParsedProperty | ParsedComponent] = []
        for field_name, value in values.items():
            if field_name in all_fields:
                continue
            for prop in value:
                if isinstance(prop, ParsedProperty):
                    extras.append(prop)
        if extras:
            values["extras"] = extras
        return values

    @root_validator(pre=True, allow_reuse=True)
    def parse_property_values(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Parse individual ParsedProperty value fields."""
        _LOGGER.debug("Parsing value data %s", values)

        for field in cls.__fields__.values():
            if field.alias == "extras":
                continue
            if not (value := values.get(field.alias)):
                continue
            if not (isinstance(value, list) and isinstance(value[0], ParsedProperty)):
                # The incoming value is not from the parse tree
                continue
            if field.alias in EXPAND_REPEATED_VALUES:
                value = cls._expand_repeated_property(value)
            # Repeated values will accept a list, otherwise truncate to a single
            # value when repeated is not allowed.
            allow_repeated = field.shape == SHAPE_LIST
            if not allow_repeated and len(value) > 1:
                raise ValueError(f"Expected one value for field: {field.alias}")
            field_types = cls._get_field_types(field.type_)
            validated = [cls._parse_property(field_types, prop) for prop in value]
            values[field.alias] = validated if allow_repeated else validated[0]

        _LOGGER.debug("Completed parsing value data %s", values)

        return values

    @classmethod
    def _parse_property(cls, field_types: list[type], prop: ParsedProperty) -> Any:
        """Parse an individual field value from a ParsedProperty as the specified types."""
        _LOGGER.debug(
            "Parsing field '%s' with value '%s' as types %s",
            prop.name,
            prop.value,
            field_types,
        )
        errors = []
        for sub_type in field_types:
            try:
                return cls._parse_single_property(sub_type, prop)
            except ValueError as err:
                _LOGGER.debug(
                    "Unable to parse property value as type %s: %s", sub_type, err
                )
                errors.append(str(err))
                continue
        raise ValueError(f"Failed to validate: {prop.value}, errors: ({errors})")

    @classmethod
    def _parse_single_property(cls, field_type: type, prop: ParsedProperty) -> Any:
        """Parse an individual field as a single type."""
        if (
            value_type := prop.get_parameter_value(ATTR_VALUE)
        ) and field_type not in DATA_TYPE.disable_value_param:
            # Property parameter specified a strong type
            if func := DATA_TYPE.parse_parameter_by_name.get(value_type):
                _LOGGER.debug("Parsing %s as value type '%s'", prop.name, value_type)
                return func(prop)
            # Consider graceful degradation instead in the future
            raise ValueError(
                f"Property parameter specified unsupported type: {value_type}"
            )

        if decoder := DATA_TYPE.parse_property_value.get(field_type):
            _LOGGER.debug("Decoding '%s' as type '%s'", prop.name, field_type)
            return decoder(prop)

        _LOGGER.debug("Using '%s' bare property value '%s'", prop.name, prop.value)
        return prop.value

    @classmethod
    def _expand_repeated_property(
        cls, value: list[ParsedProperty]
    ) -> list[ParsedProperty]:
        """Expand properties with repeated values into separate properties."""
        result: list[ParsedProperty] = []
        for prop in value:
            if "," in prop.value:
                for sub_value in prop.value.split(","):
                    sub_prop = copy.deepcopy(prop)
                    sub_prop.value = sub_value
                    result.append(sub_prop)
            else:
                result.append(prop)
        return result

    @classmethod
    def _get_field_types(cls, field_type: type) -> list[type]:
        """Return type to attempt for encoding/decoding based on the field type."""
        origin = get_origin(field_type)
        if origin is Union:
            if not (args := get_args(field_type)):
                raise ValueError(f"Unable to determine args of type: {field_type}")

            # get_args does not have a deterministic order, so use the order supplied
            # in the registry. Ignore None as its not a parseable type.
            sortable_args = [
                (DATA_TYPE.parse_order.get(arg, 0), arg)
                for arg in args
                if arg is not type(None)  # noqa: E721
            ]
            sortable_args.sort(reverse=True)
            return [arg for (order, arg) in sortable_args]
        return [field_type]

    def __encode_component_root__(self) -> ParsedComponent:
        """Encode the calendar stream as an rfc5545 iCalendar content."""
        # The overall data model hierarchy is created by pydantic and properties
        # are encoded using the json encoders specific for each type. These are
        # marshalled through as string values. There are then additional passes
        # to ge the data in to the right final format for ics encoding.
        model_data = json.loads(
            self.json(by_alias=True, exclude_none=True, exclude_defaults=True)
        )
        # The component name is ignored as we're really only encoding children components
        return self.__encode_component__(self.__class__.__name__, model_data)

    @classmethod
    def __encode_component__(
        cls, name: str, model_data: dict[str, Any]
    ) -> ParsedComponent:
        """Encode this object as a component to prepare for serialization.

        The data passed in have already been encoded with one pass from the root json
        encoder. This method takes additional passes to add more field specific encoding,
        as well as overall component objects.
        """
        parent = ParsedComponent(name=name)
        for field in cls.__fields__.values():
            key = field.alias
            values = model_data.get(key)
            if values is None or key == "extras":
                continue
            if not isinstance(values, list):
                values = [values]
            for value in values:
                if component_encoder := getattr(
                    field.type_, "__encode_component__", None
                ):
                    parent.components.append(component_encoder(key, value))
                    continue
                if prop := cls._encode_property(key, field.type_, value):
                    parent.properties.append(prop)
        return parent

    @classmethod
    def _encode_property(cls, key: str, field_type: type, value: Any) -> ParsedProperty:
        """Encode an individual property for the specified field."""
        # A property field may have multiple possible types, like for
        # a Union. Pick the first type that is able to encode the value.
        errors = []
        for sub_type in cls._get_field_types(field_type):
            encoded_value: Any | None = None
            if value_encoder := DATA_TYPE.encode_property_value.get(sub_type):
                try:
                    encoded_value = value_encoder(value)
                except ValueError as err:
                    _LOGGER.debug("Encoding failed for property: %s", err)
                    errors.append(str(err))
                    continue
            else:
                encoded_value = value

            if encoded_value is not None:
                prop = ParsedProperty(name=key, value=encoded_value)
                if params_encoder := DATA_TYPE.encode_property_params.get(
                    sub_type, None
                ):
                    if params := params_encoder(value):
                        prop.params = params
                return prop

        raise ValueError(f"Unable to encode property: {value}, errors: {errors}")

    class Config:
        """Pyandtic model configuration."""

        validate_assignment = True
        allow_population_by_field_name = True
        smart_union = True
        arbitrary_types_allowed = True



"""A grouping of component properties for describing recurrence rules.

This is a standalone component used for defining recurrence rules for
any other component. This is a faster performance library than invoking
the full rfc5545 parser for the same data. As a result, it does not support
the full rfc5545 specification, but only a subset of it.
"""

from __future__ import annotations

import datetime
import logging
from collections.abc import Iterable
from typing import Any, Optional, Union, Self

from dateutil import rrule
from pydantic.v1 import BaseModel, Field

from .parsing.property import (
    ParsedProperty,
    ParsedPropertyParameter,
    parse_basic_ics_properties,
)
from .parsing.component import ParsedComponent

from .types.data_types import DATA_TYPE
from .types.date import DateEncoder
from .types.recur import Recur
from .types.date_time import DateTimeEncoder
from .component import ComponentModel
from .exceptions import CalendarParseError
from .iter import RulesetIterable


_LOGGER = logging.getLogger(__name__)


class Recurrences(ComponentModel):
    """A common set of recurrence related properties for calendar components."""

    dtstart: Optional[Union[datetime.datetime | datetime.date]] = None
    """The start date for the event."""

    rrule: list[Recur] = Field(default_factory=list)
    """The recurrence rule for the event."""

    rdate: list[Union[datetime.datetime, datetime.date]] = Field(default_factory=list)
    """Dates for the event."""

    exdate: list[Union[datetime.datetime, datetime.date]] = Field(default_factory=list)
    """Excluded dates for the event."""

    @classmethod
    def from_basic_contentlines(cls, contentlines: list[str]) -> Self:
        """Parse a raw component from a list of contentlines making up the body.

        This is meant to be a performance optimized version of the parsing
        that is done in the main parser. It is used for parsing recurrences
        from a calendar component only.
        """
        try:
            properties = list(parse_basic_ics_properties(contentlines))
        except ValueError as err:
            raise CalendarParseError(f"Failed to parse recurrence: {err}") from err
        component = ParsedComponent(
            name="recurrences",  # Not used in the model
            properties=properties,
        )
        return cls.parse_obj(component.as_dict())

    def as_rrule(
        self, dtstart: datetime.date | datetime.datetime | None = None
    ) -> Iterable[datetime.date | datetime.datetime]:
        """Return the set of recurrences as a rrule that emits start times."""
        if dtstart is None:
            dtstart = self.dtstart
        if dtstart is None:
            raise ValueError("dtstart is required to generate recurrences")
        return RulesetIterable(
            dtstart,
            [rule.as_rrule(dtstart) for rule in self.rrule],
            self.rdate,
            self.exdate,
        )

    def ics(self) -> list[str]:
        """Serialize the recurrence rules as strings."""
        return [prop.ics() for prop in self.__encode_component_root__().properties]

    class Config:
        """Configuration for IcsCalendarStream pydantic model."""

        json_encoders = DATA_TYPE.encode_property_json
        validate_assignment = True
        allow_population_by_field_name = True



"""Tests for rfc8536 examples."""

import io
import json
import pathlib
import re

import pytest
from syrupy import SnapshotAssertion

from ical.tzif.tzif import read_tzif

RFC_LINE = re.compile(r"\|(?:[0-9]|\s)+\| (.*?)\s+\| .+ \| .+ \|")

TESTDATA_PATH = pathlib.Path("tests/tzif/testdata/")
TESTDATA_FILES = list(TESTDATA_PATH.glob("*.txt"))
TESTDATA_IDS = [x.stem for x in TESTDATA_FILES]


def rfc_to_binary(rfc_text: str) -> bytes:
    """Convert the RFC example text to a binary blob."""
    buf = io.BytesIO()
    for line in rfc_text.split("\n"):
        match = RFC_LINE.match(line)
        if not match:
            continue
        if not (payload := match.group(1)):
            continue
        buf.write(bytearray.fromhex(payload.replace(" ", "")))
    return buf.getvalue()


@pytest.mark.parametrize("filename", TESTDATA_FILES, ids=TESTDATA_IDS)
def test_parse(
    filename: pathlib.Path, snapshot: SnapshotAssertion, json_encoder: json.JSONEncoder
) -> None:
    """Test that reads RFC examples from golden files."""
    content = rfc_to_binary(filename.read_text())
    print(content)
    result = read_tzif(content)
    obj_data = json.loads(json.dumps(result, default=json_encoder.default))
    assert obj_data == snapshot



"""Tests for DURATION data types."""

import datetime

import pytest

from ical.component import ComponentModel
from ical.parsing.property import ParsedProperty
from ical.types.data_types import DATA_TYPE


class FakeModel(ComponentModel):
    """Model under test."""

    duration: datetime.timedelta

    class Config:
        """Pydantic model configuration."""

        json_encoders = DATA_TYPE.encode_property_json


@pytest.mark.parametrize(
    "value,duration,encoded_value",
    [
        (
            "P15DT5H0M20S",
            datetime.timedelta(days=15, hours=5, seconds=20),
            "P2W1DT5H20S",
        ),
        ("P7W", datetime.timedelta(days=7 * 7), "P7W"),
        ("-P7W", datetime.timedelta(days=-7 * 7), "-P7W"),
    ],
)
def test_duration(value: str, duration: datetime.timedelta, encoded_value: str) -> None:
    """Test for duration fields."""

    model = FakeModel.parse_obj(
        {"duration": [ParsedProperty(name="duration", value=value)]}
    )
    assert model.duration == duration
    component = model.__encode_component_root__()
    assert component.name == "FakeModel"
    assert component.properties == [
        ParsedProperty(name="duration", value=encoded_value)
    ]


def test_duration_from_object() -> None:
    """Test for a duration field from a native object."""

    model = FakeModel(duration=datetime.timedelta(hours=1))
    assert model.duration == datetime.timedelta(hours=1)

    component = model.__encode_component_root__()
    assert component.name == "FakeModel"
    assert component.properties == [ParsedProperty(name="duration", value="PT1H")]



"""Tests for BOOLEAN data types."""

import pytest

from ical.component import ComponentModel
from ical.exceptions import CalendarParseError
from ical.parsing.property import ParsedProperty


class FakeModel(ComponentModel):
    """Model under test."""

    example: bool


def test_bool() -> None:
    """Test for boolean fields."""

    model = FakeModel.parse_obj(
        {"example": [ParsedProperty(name="example", value="TRUE")]}
    )
    assert model.example

    model = FakeModel.parse_obj(
        {"example": [ParsedProperty(name="example", value="FALSE")]}
    )
    assert not model.example

    with pytest.raises(CalendarParseError):
        FakeModel.parse_obj({"example": [ParsedProperty(name="example", value="efd")]})

    # Populate based on bool object
    model = FakeModel(example=True)
    assert model.example
    component = model.__encode_component_root__()
    assert component.properties == [
        ParsedProperty(name="example", value="TRUE"),
    ]

    model = FakeModel(example=False)
    assert not model.example
    component = model.__encode_component_root__()
    assert component.properties == [
        ParsedProperty(name="example", value="FALSE"),
    ]



"""Tests for timeline related calendar eents."""
from __future__ import annotations

import datetime
import zoneinfo

import pytest
from ical.exceptions import CalendarParseError

from ical.calendar import Calendar
from ical.component import ComponentModel
from ical.event import Event
from ical.parsing.property import ParsedProperty, ParsedPropertyParameter
from ical.timeline import Timeline
from ical.todo import Todo
from ical.types.recur import Frequency, Recur, RecurrenceId, Weekday, WeekdayValue


def recur_timeline(event: Event) -> Timeline:
    """Create a timeline for the specified recurring event."""
    calendar = Calendar()
    calendar.events.append(event)
    return calendar.timeline


class FakeModel(ComponentModel):
    """Model under test."""

    recurrence_id: RecurrenceId


def test_recurrence_id_datetime() -> None:
    """Test a recurrence id datetime field."""

    model = FakeModel.parse_obj(
        {
            "recurrence_id": [
                ParsedProperty(name="recurrence_id", value="20220724T120000")
            ]
        }
    )
    assert model.recurrence_id == "20220724T120000"


def test_recurrence_id_date() -> None:
    """Test a recurrence id date field."""

    model = FakeModel.parse_obj(
        {"recurrence_id": [ParsedProperty(name="recurrence_id", value="20220724")]}
    )
    assert model.recurrence_id == "20220724"


def test_recurrence_id_ignore_params() -> None:
    """Test property parameter values are ignored."""

    model = FakeModel.parse_obj(
        {
            "recurrence_id": [
                ParsedProperty(
                    name="recurrence_id",
                    value="20220724T120000",
                    params=[
                        ParsedPropertyParameter(name="RANGE", values=["THISANDFUTURE"]),
                    ],
                )
            ]
        }
    )
    assert model.recurrence_id == "20220724T120000"


def test_invalid_recurrence_id() -> None:
    """Test for a recurrence id that is not a valid DATE or DATE-TIME string."""
    model = FakeModel.parse_obj(
        {"recurrence_id": [ParsedProperty(name="recurrence_id", value="invalid")]}
    )
    assert model.recurrence_id == "invalid"


@pytest.mark.parametrize(
    "start,end,rrule,expected",
    [
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(freq=Frequency.DAILY, until=datetime.date(2022, 8, 4)),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 8, 2), datetime.date(2022, 8, 3)),
                (datetime.date(2022, 8, 3), datetime.date(2022, 8, 4)),
                (datetime.date(2022, 8, 4), datetime.date(2022, 8, 5)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(freq=Frequency.DAILY, until=datetime.date(2022, 8, 4), interval=2),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 8, 3), datetime.date(2022, 8, 4)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(freq=Frequency.DAILY, count=3),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 8, 2), datetime.date(2022, 8, 3)),
                (datetime.date(2022, 8, 3), datetime.date(2022, 8, 4)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(freq=Frequency.DAILY, interval=2, count=3),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 8, 3), datetime.date(2022, 8, 4)),
                (datetime.date(2022, 8, 5), datetime.date(2022, 8, 6)),
            ],
        ),
        (
            datetime.datetime(2022, 8, 1, 9, 30, 0),
            datetime.datetime(2022, 8, 1, 10, 0, 0),
            Recur(freq=Frequency.DAILY, until=datetime.datetime(2022, 8, 4, 9, 30, 0)),
            [
                (
                    datetime.datetime(2022, 8, 1, 9, 30, 0),
                    datetime.datetime(2022, 8, 1, 10, 0, 0),
                ),
                (
                    datetime.datetime(2022, 8, 2, 9, 30, 0),
                    datetime.datetime(2022, 8, 2, 10, 0, 0),
                ),
                (
                    datetime.datetime(2022, 8, 3, 9, 30, 0),
                    datetime.datetime(2022, 8, 3, 10, 0, 0),
                ),
                (
                    datetime.datetime(2022, 8, 4, 9, 30, 0),
                    datetime.datetime(2022, 8, 4, 10, 0, 0),
                ),
            ],
        ),
    ],
)
def test_day_iteration(
    start: datetime.datetime | datetime.date,
    end: datetime.datetime | datetime.date,
    rrule: Recur,
    expected: list[tuple[datetime.date, datetime.date]],
) -> None:
    """Test recurrence rules for day frequency."""
    event = Event(
        summary="summary",
        start=start,
        end=end,
        rrule=rrule,
    )
    timeline = recur_timeline(event)
    assert [(e.start, e.end) for e in timeline] == expected


@pytest.mark.parametrize(
    "start,end,rrule,expected",
    [
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(
                freq=Frequency.WEEKLY,
                until=datetime.date(2022, 9, 6),
                by_weekday=[WeekdayValue(Weekday.MONDAY)],
            ),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 8, 8), datetime.date(2022, 8, 9)),
                (datetime.date(2022, 8, 15), datetime.date(2022, 8, 16)),
                (datetime.date(2022, 8, 22), datetime.date(2022, 8, 23)),
                (datetime.date(2022, 8, 29), datetime.date(2022, 8, 30)),
                (datetime.date(2022, 9, 5), datetime.date(2022, 9, 6)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(
                freq=Frequency.WEEKLY,
                until=datetime.date(2022, 9, 6),
                interval=2,
                by_weekday=[WeekdayValue(Weekday.MONDAY)],
            ),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 8, 15), datetime.date(2022, 8, 16)),
                (datetime.date(2022, 8, 29), datetime.date(2022, 8, 30)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(
                freq=Frequency.WEEKLY,
                count=3,
                by_weekday=[WeekdayValue(Weekday.MONDAY)],
            ),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 8, 8), datetime.date(2022, 8, 9)),
                (datetime.date(2022, 8, 15), datetime.date(2022, 8, 16)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(
                freq=Frequency.WEEKLY,
                interval=2,
                count=3,
                by_weekday=[WeekdayValue(Weekday.MONDAY)],
            ),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 8, 15), datetime.date(2022, 8, 16)),
                (datetime.date(2022, 8, 29), datetime.date(2022, 8, 30)),
            ],
        ),
    ],
)
def test_weekly_iteration(
    start: datetime.date | datetime.date,
    end: datetime.date | datetime.date,
    rrule: Recur,
    expected: list[tuple[datetime.date, datetime.date]],
) -> None:
    """Test recurrence rules for weekly frequency."""
    event = Event(
        summary="summary",
        start=start,
        end=end,
        rrule=rrule,
    )
    timeline = recur_timeline(event)
    assert [(e.start, e.end) for e in timeline] == expected


@pytest.mark.parametrize(
    "start,end,rrule,expected",
    [
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(
                freq=Frequency.MONTHLY,
                until=datetime.date(2023, 1, 1),
                by_month_day=[1],
            ),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 9, 1), datetime.date(2022, 9, 2)),
                (datetime.date(2022, 10, 1), datetime.date(2022, 10, 2)),
                (datetime.date(2022, 11, 1), datetime.date(2022, 11, 2)),
                (datetime.date(2022, 12, 1), datetime.date(2022, 12, 2)),
                (datetime.date(2023, 1, 1), datetime.date(2023, 1, 2)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(
                freq=Frequency.MONTHLY,
                until=datetime.date(2023, 1, 1),
                interval=2,
                by_month_day=[1],
            ),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 10, 1), datetime.date(2022, 10, 2)),
                (datetime.date(2022, 12, 1), datetime.date(2022, 12, 2)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(freq=Frequency.MONTHLY, count=3, by_month_day=[1]),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 9, 1), datetime.date(2022, 9, 2)),
                (datetime.date(2022, 10, 1), datetime.date(2022, 10, 2)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur(
                freq=Frequency.MONTHLY,
                interval=2,
                count=3,
                by_month_day=[1],
            ),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 10, 1), datetime.date(2022, 10, 2)),
                (datetime.date(2022, 12, 1), datetime.date(2022, 12, 2)),
            ],
        ),
        (
            datetime.datetime(2022, 8, 2, 9, 0, 0),
            datetime.datetime(2022, 8, 2, 9, 30, 0),
            Recur(
                freq=Frequency.MONTHLY,
                until=datetime.datetime(2023, 1, 1, 0, 0, 0),
                by_month_day=[2],
            ),
            [
                (
                    datetime.datetime(2022, 8, 2, 9, 0, 0),
                    datetime.datetime(2022, 8, 2, 9, 30, 0),
                ),
                (
                    datetime.datetime(2022, 9, 2, 9, 0, 0),
                    datetime.datetime(2022, 9, 2, 9, 30, 0),
                ),
                (
                    datetime.datetime(2022, 10, 2, 9, 0, 0),
                    datetime.datetime(2022, 10, 2, 9, 30, 0),
                ),
                (
                    datetime.datetime(2022, 11, 2, 9, 0, 0),
                    datetime.datetime(2022, 11, 2, 9, 30, 0),
                ),
                (
                    datetime.datetime(2022, 12, 2, 9, 0, 0),
                    datetime.datetime(2022, 12, 2, 9, 30, 0),
                ),
            ],
        ),
        (
            datetime.date(2022, 8, 7),
            datetime.date(2022, 8, 8),
            Recur(
                freq=Frequency.MONTHLY,
                interval=2,
                count=3,
                by_weekday=[WeekdayValue(weekday=Weekday.SUNDAY, occurrence=1)],
            ),
            [
                (datetime.date(2022, 8, 7), datetime.date(2022, 8, 8)),
                (datetime.date(2022, 10, 2), datetime.date(2022, 10, 3)),
                (datetime.date(2022, 12, 4), datetime.date(2022, 12, 5)),
            ],
        ),
        (
            datetime.date(2022, 8, 7),
            datetime.date(2022, 8, 8),
            Recur(
                freq=Frequency.MONTHLY,
                count=3,
                by_weekday=[WeekdayValue(weekday=Weekday.SUNDAY, occurrence=-1)],
            ),
            [
                (datetime.date(2022, 8, 28), datetime.date(2022, 8, 29)),
                (datetime.date(2022, 9, 25), datetime.date(2022, 9, 26)),
                (datetime.date(2022, 10, 30), datetime.date(2022, 10, 31)),
            ],
        ),
        (
            datetime.date(2022, 8, 1),
            datetime.date(2022, 8, 2),
            Recur.from_rrule("FREQ=MONTHLY;INTERVAL=2;COUNT=3;BYMONTHDAY=1"),
            [
                (datetime.date(2022, 8, 1), datetime.date(2022, 8, 2)),
                (datetime.date(2022, 10, 1), datetime.date(2022, 10, 2)),
                (datetime.date(2022, 12, 1), datetime.date(2022, 12, 2)),
            ],
        ),
    ],
)
def test_monthly_iteration(
    start: datetime.date | datetime.date,
    end: datetime.date | datetime.date,
    rrule: Recur,
    expected: list[tuple[datetime.date, datetime.date]],
) -> None:
    """Test recurrency rules for monthly frequency."""
    event = Event(
        summary="summary",
        start=start,
        end=end,
        rrule=rrule,
    )
    timeline = recur_timeline(event)
    assert [(e.start, e.end) for e in timeline] == expected


def test_recur_no_bound() -> None:
    """Test a recurrence rule with no end date."""

    event = Event(
        summary="summary",
        start=datetime.date(2022, 8, 1),
        end=datetime.date(2022, 8, 2),
        rrule=Recur(freq=Frequency.DAILY, interval=2),
    )
    timeline = recur_timeline(event)

    def on_date(day: datetime.date) -> list[datetime.date]:
        return [e.start for e in timeline.on_date(day)]

    assert on_date(datetime.date(2022, 8, 1)) == [datetime.date(2022, 8, 1)]
    assert on_date(datetime.date(2022, 8, 2)) == []
    assert on_date(datetime.date(2022, 8, 3)) == [datetime.date(2022, 8, 3)]

    assert on_date(datetime.date(2025, 1, 1)) == [datetime.date(2025, 1, 1)]
    assert on_date(datetime.date(2025, 1, 2)) == []
    assert on_date(datetime.date(2025, 1, 3)) == [datetime.date(2025, 1, 3)]

    assert on_date(datetime.date(2035, 9, 1)) == []
    assert on_date(datetime.date(2035, 9, 2)) == [datetime.date(2035, 9, 2)]
    assert on_date(datetime.date(2035, 9, 3)) == []


def test_merged_recur_event_timeline() -> None:
    """Test merged recurring and events timeline."""
    calendar = Calendar()
    calendar.events.extend(
        [
            Event(
                summary="Morning exercise",
                start=datetime.datetime(2022, 8, 2, 6, 0, 0),
                end=datetime.datetime(2022, 8, 2, 7, 0, 0),
                rrule=Recur(freq=Frequency.DAILY),
            ),
            Event(
                summary="Meeting",
                start=datetime.datetime(2022, 8, 2, 10, 0, 0),
                end=datetime.datetime(2022, 8, 2, 10, 30, 0),
            ),
            Event(
                summary="Trash day",
                start=datetime.date(2022, 8, 3),
                end=datetime.date(2022, 8, 4),
                rrule=Recur(
                    freq=Frequency.WEEKLY, by_weekday=[WeekdayValue(Weekday.WEDNESDAY)]
                ),
            ),
            Event(
                summary="Appointment",
                start=datetime.datetime(2022, 8, 5, 8, 0, 0),
                end=datetime.datetime(2022, 8, 5, 8, 30, 0),
            ),
            Event(
                summary="Pay day",
                start=datetime.date(2022, 8, 15),
                end=datetime.date(2022, 8, 16),
                rrule=Recur(freq=Frequency.MONTHLY, by_month_day=[15]),
            ),
        ]
    )
    events = calendar.timeline.included(
        datetime.date(2022, 8, 1),
        datetime.date(2022, 9, 2),
    )
    assert [(event.start, event.summary) for event in events] == [
        (datetime.datetime(2022, 8, 2, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 2, 10, 0, 0), "Meeting"),
        (datetime.date(2022, 8, 3), "Trash day"),
        (datetime.datetime(2022, 8, 3, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 4, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 5, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 5, 8, 0, 0), "Appointment"),
        (datetime.datetime(2022, 8, 6, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 7, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 8, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 9, 6, 0, 0), "Morning exercise"),
        (datetime.date(2022, 8, 10), "Trash day"),
        (datetime.datetime(2022, 8, 10, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 11, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 12, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 13, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 14, 6, 0, 0), "Morning exercise"),
        (datetime.date(2022, 8, 15), "Pay day"),
        (datetime.datetime(2022, 8, 15, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 16, 6, 0, 0), "Morning exercise"),
        (datetime.date(2022, 8, 17), "Trash day"),
        (datetime.datetime(2022, 8, 17, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 18, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 19, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 20, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 21, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 22, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 23, 6, 0, 0), "Morning exercise"),
        (datetime.date(2022, 8, 24), "Trash day"),
        (datetime.datetime(2022, 8, 24, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 25, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 26, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 27, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 28, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 29, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 30, 6, 0, 0), "Morning exercise"),
        (datetime.date(2022, 8, 31), "Trash day"),
        (datetime.datetime(2022, 8, 31, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 9, 1, 6, 0, 0), "Morning exercise"),
    ]


def test_exclude_date() -> None:
    """Test recurrence rule with date exclusions."""
    calendar = Calendar()
    calendar.events.extend(
        [
            Event(
                summary="Morning exercise",
                start=datetime.datetime(2022, 8, 2, 6, 0, 0),
                end=datetime.datetime(2022, 8, 2, 7, 0, 0),
                rrule=Recur(
                    freq=Frequency.DAILY, until=datetime.datetime(2022, 8, 10, 6, 0, 0)
                ),
                exdate=[
                    datetime.datetime(2022, 8, 3, 6, 0, 0),
                    datetime.datetime(2022, 8, 4, 6, 0, 0),
                    datetime.datetime(2022, 8, 6, 6, 0, 0),
                    datetime.datetime(2022, 8, 7, 6, 0, 0),
                ],
            ),
        ]
    )
    events = list(calendar.timeline)
    assert [(event.start, event.summary) for event in events] == [
        (datetime.datetime(2022, 8, 2, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 5, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 8, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 9, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 10, 6, 0, 0), "Morning exercise"),
    ]


def test_rdate() -> None:
    """Test recurrence rule with specific dates and exclusions."""
    calendar = Calendar()
    calendar.events.extend(
        [
            Event(
                summary="Morning exercise",
                start=datetime.datetime(2022, 8, 2, 6, 0, 0),
                end=datetime.datetime(2022, 8, 2, 7, 0, 0),
                rdate=[
                    datetime.datetime(2022, 8, 2, 6, 0, 0),
                    datetime.datetime(2022, 8, 3, 6, 0, 0),
                    datetime.datetime(2022, 8, 4, 6, 0, 0),
                    datetime.datetime(2022, 8, 5, 6, 0, 0),
                    datetime.datetime(2022, 8, 6, 6, 0, 0),
                    datetime.datetime(2022, 8, 7, 6, 0, 0),
                    datetime.datetime(2022, 8, 8, 6, 0, 0),
                    datetime.datetime(2022, 8, 9, 6, 0, 0),
                    datetime.datetime(2022, 8, 10, 6, 0, 0),
                ],
                exdate=[
                    datetime.datetime(2022, 8, 3, 6, 0, 0),
                    datetime.datetime(2022, 8, 4, 6, 0, 0),
                    datetime.datetime(2022, 8, 6, 6, 0, 0),
                    datetime.datetime(2022, 8, 7, 6, 0, 0),
                ],
            ),
        ]
    )
    events = list(calendar.timeline)
    assert [(event.start, event.summary) for event in events] == [
        (datetime.datetime(2022, 8, 2, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 5, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 8, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 9, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 10, 6, 0, 0), "Morning exercise"),
    ]


def test_year_iteration() -> None:
    """Test iteration with a yearly frequency."""
    calendar = Calendar()
    calendar.events.extend(
        [
            Event(
                summary="Bi-annual meeting",
                start=datetime.datetime(2022, 1, 2, 6, 0, 0),
                end=datetime.datetime(2022, 1, 2, 7, 0, 0),
                rrule=Recur(freq=Frequency.YEARLY, by_month=[1, 6], count=4),
            ),
        ]
    )
    events = list(calendar.timeline)
    assert [(event.start, event.summary) for event in events] == [
        (datetime.datetime(2022, 1, 2, 6, 0, 0), "Bi-annual meeting"),
        (datetime.datetime(2022, 6, 2, 6, 0, 0), "Bi-annual meeting"),
        (datetime.datetime(2023, 1, 2, 6, 0, 0), "Bi-annual meeting"),
        (datetime.datetime(2023, 6, 2, 6, 0, 0), "Bi-annual meeting"),
    ]


@pytest.mark.parametrize(
    ("tzinfo", "until_tzinfo"),
    [
        (None, None),
        (zoneinfo.ZoneInfo("America/New_York"), datetime.timezone.utc),
    ],
)
def test_until_time_valid(
    tzinfo: zoneinfo.ZoneInfo | None, until_tzinfo: zoneinfo.ZoneInfo | None
) -> None:
    """Test success cases where until has a valid date or time compared to dtstart."""

    Event(
        summary="Bi-annual meeting",
        start=datetime.datetime(2022, 1, 2, 6, 0, 0, tzinfo=tzinfo),
        end=datetime.datetime(2022, 1, 2, 7, 0, 0, tzinfo=tzinfo),
        rrule=Recur(
            freq=Frequency.DAILY,
            until=datetime.datetime(2022, 8, 4, 6, 0, 0, tzinfo=until_tzinfo),
        ),
    )


def test_until_time_mismatch() -> None:
    """Test failure case where until has a different timezone than start."""

    with pytest.raises(
        CalendarParseError,
        match="DTSTART was DATE-TIME but UNTIL was DATE",
    ):
        Event(
            summary="Bi-annual meeting",
            start=datetime.datetime(2022, 1, 2, 6, 0, 0),
            end=datetime.datetime(2022, 1, 2, 7, 0, 0),
            rrule=Recur(
                freq=Frequency.DAILY,
                until=datetime.date(2022, 8, 4),
            ),
        )

    with pytest.raises(
        CalendarParseError, match="DTSTART is date local but UNTIL was not"
    ):
        Event(
            summary="Bi-annual meeting",
            start=datetime.datetime(2022, 1, 2, 6, 0, 0),
            end=datetime.datetime(2022, 1, 2, 7, 0, 0),
            rrule=Recur(
                freq=Frequency.DAILY,
                until=datetime.datetime(
                    2022, 8, 4, 1, 0, 0, tzinfo=datetime.timezone.utc
                ),
            ),
        )

    with pytest.raises(
        CalendarParseError, match="DTSTART had UTC or local and UNTIL must be UTC"
    ):
        Event(
            summary="Bi-annual meeting",
            start=datetime.datetime(2022, 1, 2, 6, 0, 0, tzinfo=datetime.timezone.utc),
            end=datetime.datetime(2022, 1, 2, 7, 0, 0),
            rrule=Recur(
                freq=Frequency.DAILY,
                until=datetime.datetime(
                    2022, 8, 4, 1, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/New_York")
                ),
            ),
        )

    with pytest.raises(
        CalendarParseError, match="DTSTART had UTC or local and UNTIL must be UTC"
    ):
        Event(
            summary="Bi-annual meeting",
            start=datetime.datetime(
                2022, 1, 2, 6, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/New_York")
            ),
            end=datetime.datetime(2022, 1, 2, 7, 0, 0),
            rrule=Recur(
                freq=Frequency.DAILY,
                until=datetime.datetime(
                    2022, 8, 4, 1, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/New_York")
                ),
            ),
        )


@pytest.mark.parametrize(
    "recur",
    [
        Recur(freq=Frequency.DAILY, interval=2),
        Recur.from_rrule("FREQ=DAILY;INTERVAL=2"),
    ],
)
def test_recur_as_string(recur: Recur) -> None:
    """Test converting a recurrence rule back to a string."""

    event = Event(
        summary="summary",
        start=datetime.date(2022, 8, 1),
        end=datetime.date(2022, 8, 2),
        rrule=recur,
    )
    assert event.rrule
    assert event.rrule.as_rrule_str() == "FREQ=DAILY;INTERVAL=2"


@pytest.mark.parametrize(
    "recur",
    [
        Recur(freq=Frequency.DAILY, interval=2),
        Recur.from_rrule("FREQ=DAILY;INTERVAL=2"),
    ],
)
def test_todo_recur_as_string(recur: Recur) -> None:
    """Test converting a recurrence rule back to a string."""

    event = Todo(
        summary="summary",
        due=datetime.date(2022, 8, 1),
        rrule=recur,
    )
    assert event.rrule
    assert event.rrule.as_rrule_str() == "FREQ=DAILY;INTERVAL=2"


@pytest.mark.parametrize(
    "recur",
    [
        Recur(freq=Frequency.DAILY, until=datetime.datetime(2022, 8, 10, 6, 0, 0)),
        Recur.from_rrule("FREQ=DAILY;UNTIL=20220810T060000"),
    ],
)
def test_recur_until_as_string(recur: Recur) -> None:
    """Test converting a recurrence rule with a date into a string."""

    event = Event(
        summary="summary",
        start=datetime.datetime(2022, 8, 1, 6, 0, 0),
        end=datetime.datetime(2022, 8, 2, 6, 30, 0),
        rrule=recur,
    )
    assert event.rrule
    assert event.rrule.as_rrule_str() == "FREQ=DAILY;UNTIL=20220810T060000"


@pytest.mark.parametrize(
    "recur",
    [
        Recur(freq=Frequency.WEEKLY, by_weekday=[WeekdayValue(Weekday.TUESDAY)]),
        Recur.from_rrule("FREQ=WEEKLY;BYDAY=TU"),
    ],
)
def test_recur_by_weekday_as_string(recur: Recur) -> None:
    """Test converting a recurrence rule with a weekday repeat into a string."""
    event = Event(
        summary="summary",
        start=datetime.date(2022, 9, 6),
        end=datetime.date(2022, 9, 7),
        rrule=recur,
    )
    assert event.rrule
    assert event.rrule.as_rrule_str() == "FREQ=WEEKLY;BYDAY=TU"


@pytest.mark.parametrize(
    "recur",
    [
        Recur(
            freq=Frequency.MONTHLY, until=datetime.date(2023, 1, 1), by_month_day=[1]
        ),
        Recur.from_rrule("FREQ=MONTHLY;UNTIL=20230101;BYMONTHDAY=1"),
    ],
)
def test_recur_by_monthday_as_string(recur: Recur) -> None:
    """Test converting a recurrence rule with a weekday reepat into a string."""
    event = Event(
        summary="summary",
        start=datetime.date(2022, 9, 6),
        end=datetime.date(2022, 9, 7),
        rrule=recur,
    )
    assert event.rrule
    assert event.rrule.as_rrule_str() == "FREQ=MONTHLY;UNTIL=20230101;BYMONTHDAY=1"


@pytest.mark.parametrize(
    "recur",
    [
        Recur(
            freq=Frequency.MONTHLY,
            by_weekday=[WeekdayValue(weekday=Weekday.TUESDAY, occurrence=-1)],
        ),
        Recur.from_rrule("FREQ=MONTHLY;BYDAY=-1TU"),
    ],
)
def test_recur_by_last_day_as_string(recur: Recur) -> None:
    """Test converting a recurrence rule with a weekday repeat into a string."""
    event = Event(
        summary="summary",
        start=datetime.date(2022, 9, 6),
        end=datetime.date(2022, 9, 7),
        rrule=recur,
    )
    assert event.rrule
    assert event.rrule.as_rrule_str() == "FREQ=MONTHLY;BYDAY=-1TU"


def test_rdate_all_day() -> None:
    """Test how all day events are handled with RDATE."""

    calendar = Calendar()
    calendar.events.extend(
        [
            Event(
                summary="Monthly Event",
                start=datetime.date(2022, 8, 2),
                end=datetime.date(2022, 8, 2),
                rdate=[
                    datetime.date(2022, 8, 3),
                    datetime.date(2022, 8, 4),
                    datetime.date(2022, 10, 3),
                ],
                rrule=Recur.from_rrule("FREQ=MONTHLY;COUNT=3"),
            ),
        ]
    )
    events = list(calendar.timeline)
    assert [(event.start, event.summary) for event in events] == [
        (datetime.date(2022, 8, 2), "Monthly Event"),
        (datetime.date(2022, 8, 3), "Monthly Event"),
        (datetime.date(2022, 8, 4), "Monthly Event"),
        (datetime.date(2022, 9, 2), "Monthly Event"),
        (datetime.date(2022, 10, 2), "Monthly Event"),
        (datetime.date(2022, 10, 3), "Monthly Event"),
    ]


def test_rrule_exdate_mismatch() -> None:
    """Test a recurrence rule with mixed datetime and exdate values."""
    calendar = Calendar()
    calendar.events.extend(
        [
            Event(
                summary="Morning exercise",
                start=datetime.datetime(2022, 8, 2, 6, 0, 0),
                end=datetime.datetime(2022, 8, 2, 7, 0, 0),
                rrule=Recur(
                    freq=Frequency.WEEKLY, until=datetime.datetime(2022, 8, 10, 6, 0, 0)
                ),
                exdate=[
                    datetime.date(2022, 8, 3),
                    datetime.date(2022, 8, 4),
                    datetime.date(2022, 8, 5),
                ],
            ),
        ]
    )
    events = list(calendar.timeline)
    assert [(event.start, event.summary) for event in events] == [
        (datetime.datetime(2022, 8, 2, 6, 0, 0), "Morning exercise"),
        (datetime.datetime(2022, 8, 9, 6, 0, 0), "Morning exercise"),
    ]


# BYSETPOS


def test_bysetpos() -> None:
    """Test how all day events are handled with RDATE."""

    calendar = Calendar()
    calendar.events.extend(
        [
            Event(
                summary="Monthly Event",
                start=datetime.date(2023, 1, 31),
                end=datetime.date(2023, 2, 1),
                # Last work day of the month
                rrule=Recur.from_rrule(
                    "FREQ=MONTHLY;BYSETPOS=-1;BYDAY=MO,TU,WE,TH,FR;COUNT=5"
                ),
            ),
        ]
    )
    events = list(calendar.timeline)
    assert [(event.start, event.summary) for event in events] == [
        (datetime.date(2023, 1, 31), "Monthly Event"),
        (datetime.date(2023, 2, 28), "Monthly Event"),
        (datetime.date(2023, 3, 31), "Monthly Event"),
        (datetime.date(2023, 4, 28), "Monthly Event"),
        (datetime.date(2023, 5, 31), "Monthly Event"),
    ]



"""Test fixtures."""

from collections.abc import Generator
import dataclasses
import json
from typing import Any
from unittest.mock import patch

import pytest

try:
    from pydantic.v1.json import pydantic_encoder
except ImportError:
    from pydantic.json import pydantic_encoder  # type: ignore[no-redef]


PRODID = "-//example//1.2.3"


class DataclassEncoder(json.JSONEncoder):
    """Class that can dump data classes as dict for comparison to golden."""

    def default(self, o: Any) -> Any:
        if dataclasses.is_dataclass(o):
            # Omit empty
            return {k: v for (k, v) in dataclasses.asdict(o).items() if v}
        if isinstance(o, dict):
            return {k: v for (k, v) in o.items() if v}
        return pydantic_encoder(o)


@pytest.fixture
def json_encoder() -> json.JSONEncoder:
    """Fixture that creates a json encoder."""
    return DataclassEncoder()


@pytest.fixture(autouse=True)
def mock_prodid() -> Generator[None, None, None]:
    """Mock out the prodid used in tests."""
    with patch("ical.calendar.prodid_factory", return_value=PRODID):
        yield



"""A set of properties that describes a free/busy times."""

# pylint: disable=unnecessary-lambda

from __future__ import annotations

import datetime
import logging
from typing import Any, Optional, Union

try:
    from pydantic.v1 import Field, validator
except ImportError:
    from pydantic import Field, validator  # type: ignore[assignment]

from .component import ComponentModel
from .parsing.property import ParsedProperty
from .types import CalAddress, Period, RequestStatus, Uri
from .util import dtstamp_factory, normalize_datetime, uid_factory

_LOGGER = logging.getLogger(__name__)


class FreeBusy(ComponentModel):
    """A single free/busy entry on a calendar."""

    dtstamp: Union[datetime.datetime, datetime.date] = Field(
        default_factory=lambda: dtstamp_factory()
    )
    """Last revision date."""

    uid: str = Field(default_factory=lambda: uid_factory())
    """The persistent globally unique identifier."""

    attendees: list[CalAddress] = Field(alias="attendee", default_factory=list)
    """The user who's free/busy time is represented."""

    comment: list[str] = Field(default_factory=list)
    """Non-processing information intended to provide comments to the calendar user."""

    contacts: list[str] = Field(alias="contact", default_factory=list)
    """Contact information associated with this component."""

    # Has an alias of 'start'
    dtstart: Union[datetime.datetime, datetime.date] = Field(
        default=None,
    )
    """Start of the time range covered by this component."""

    # Has an alias of 'end'
    dtend: Optional[Union[datetime.datetime, datetime.date]] = None
    """End of the time range covered by this component."""

    freebusy: list[Period] = Field(default_factory=list)
    """The free/busy intervals."""

    organizer: Optional[CalAddress] = None
    """The calendar user who requested free/busy information."""

    request_status: Optional[RequestStatus] = Field(
        alias="request-status", default_value=None
    )
    """Return code for the scheduling request."""

    sequence: Optional[int] = None
    """The revision sequence number of this calendar component."""

    url: Optional[Uri] = None
    """The URL associated with this component."""

    # Unknown or unsupported properties
    extras: list[ParsedProperty] = Field(default_factory=list)

    def __init__(self, **data: Any) -> None:
        """Initialize Event."""
        if "start" in data:
            data["dtstart"] = data.pop("start")
        if "end" in data:
            data["dtend"] = data.pop("end")
        super().__init__(**data)

    @property
    def start(self) -> datetime.datetime | datetime.date:
        """Return the start time for the event."""
        return self.dtstart

    @property
    def end(self) -> datetime.datetime | datetime.date | None:
        """Return the end time for the event."""
        return self.dtend

    @property
    def start_datetime(self) -> datetime.datetime | None:
        """Return the events start as a datetime."""
        if not self.start:
            return None
        return normalize_datetime(self.start).astimezone(tz=datetime.timezone.utc)

    @property
    def end_datetime(self) -> datetime.datetime | None:
        """Return the events end as a datetime."""
        if not self.end:
            return None
        return normalize_datetime(self.end).astimezone(tz=datetime.timezone.utc)

    @property
    def computed_duration(self) -> datetime.timedelta | None:
        """Return the event duration."""
        if not self.end:
            return None
        return self.end - self.start

    @validator("freebusy", allow_reuse=True)
    def verify_freebusy_utc(cls, values: list[Period]) -> list[Period]:
        """Validate that the free/busy periods must be in UTC."""
        _LOGGER.info("verify_freebusy_utc")
        for value in values:
            if not value.start:
                continue
            if (
                offset := value.start.utcoffset()
            ) is None or offset.total_seconds() != 0:
                raise ValueError(f"Freebusy time must be in UTC format: {value}")

        return values



"""Tests for the tzif library."""

import datetime
import zoneinfo

import pytest

from ical.tzif import timezoneinfo, tz_rule


IGNORED_TIMEZONES = {
    "Asia/Hanoi",  # Not in tzdata
}


def test_invalid_zoneinfo() -> None:
    """Verify exception handling for an invalid timezone."""

    with pytest.raises(timezoneinfo.TimezoneInfoError, match="Unable to find timezone"):
        timezoneinfo.read("invalid")


@pytest.mark.parametrize(
    "key,dtstarts,expected_tzname,expected_offset",
    [
        (
            "America/Los_Angeles",
            [
                datetime.datetime(2021, 3, 14, 1, 59, 0),
                datetime.datetime(2021, 11, 7, 2, 0, 0),
                datetime.datetime(2022, 3, 13, 1, 59, 0),
                datetime.datetime(2022, 11, 6, 2, 0, 0),
                datetime.datetime(2023, 3, 12, 1, 59, 0),
                datetime.datetime(2023, 11, 5, 2, 0, 0),
            ],
            "PST",
            datetime.timedelta(hours=-8),
        ),
        (
            "America/Los_Angeles",
            [
                datetime.datetime(2021, 3, 14, 2, 0, 0),
                datetime.datetime(2021, 11, 7, 1, 59, 0),
                datetime.datetime(2022, 3, 13, 2, 0, 0),
                datetime.datetime(2022, 11, 6, 1, 59, 0),
                datetime.datetime(2023, 3, 12, 2, 0, 0),
                datetime.datetime(2023, 11, 5, 1, 59, 0),
            ],
            "PDT",
            datetime.timedelta(hours=-7),
        ),
        (
            "Europe/Warsaw",
            [
                datetime.datetime(2021, 3, 28, 1, 59, 0),
                datetime.datetime(2021, 10, 31, 3, 0, 0),
                datetime.datetime(2022, 3, 27, 1, 59, 0),
                datetime.datetime(2022, 10, 30, 3, 0, 0),
                datetime.datetime(2023, 3, 26, 1, 59, 0),
                datetime.datetime(2023, 10, 29, 3, 0, 0),
                datetime.datetime(2024, 3, 31, 1, 59, 0),
                datetime.datetime(2024, 10, 27, 3, 0, 0),
            ],
            "CET",
            datetime.timedelta(hours=1),
        ),
        (
            "Europe/Warsaw",
            [
                datetime.datetime(2021, 3, 28, 2, 0, 0),
                datetime.datetime(2021, 10, 31, 2, 59, 0),
                datetime.datetime(2022, 3, 27, 2, 0, 0),
                datetime.datetime(2022, 10, 30, 2, 59, 0),
                datetime.datetime(2023, 3, 26, 2, 0, 0),
                datetime.datetime(2023, 10, 29, 2, 59, 0),
                datetime.datetime(2024, 3, 31, 2, 0, 0),
                datetime.datetime(2024, 10, 27, 2, 59, 0),
            ],
            "CEST",
            datetime.timedelta(hours=2),
        ),
        (
            "Asia/Tokyo",
            [
                # Fixed offset anytime of year
                datetime.datetime(2021, 1, 1, 0, 0, 0),
                datetime.datetime(2022, 3, 1, 0, 0, 0),
                datetime.datetime(2022, 6, 1, 0, 0, 0),
                datetime.datetime(2023, 7, 1, 0, 0, 0),
                datetime.datetime(2023, 12, 1, 0, 0, 0),
            ],
            "JST",
            datetime.timedelta(hours=9),
        ),
        (
            "America/St_Thomas",
            [
                # Fixed offset anytime of year
                datetime.datetime(2021, 1, 1, 0, 0, 0),
                datetime.datetime(2022, 3, 1, 0, 0, 0),
                datetime.datetime(2022, 6, 1, 0, 0, 0),
                datetime.datetime(2023, 7, 1, 0, 0, 0),
                datetime.datetime(2023, 12, 1, 0, 0, 0),
            ],
            "AST",
            datetime.timedelta(hours=-4),
        ),
    ],
)
def test_tzinfo(
    key: str,
    dtstarts: list[datetime.datetime],
    expected_tzname: str,
    expected_offset: datetime.timedelta,
) -> None:
    """Test TzInfo implementation for known date/times."""
    tz_info = timezoneinfo.read_tzinfo(key)
    for dtstart in dtstarts:
        value = dtstart.replace(tzinfo=tz_info)
        assert tz_info.tzname(value) == expected_tzname, f"For {dtstart}"
        assert tz_info.utcoffset(value) == expected_offset, f"For {dtstart}"

    assert not tz_info.utcoffset(None)
    assert not tz_info.tzname(None)
    assert not tz_info.dst(None)


def test_rrule_str() -> None:
    """Test rule implementations for std and dst."""
    result = timezoneinfo.read("America/New_York")
    assert result.rule
    assert result.rule.dst_start
    assert isinstance(result.rule.dst_start, tz_rule.RuleDate)
    assert result.rule.dst_start.rrule_str == "FREQ=YEARLY;BYMONTH=3;BYDAY=2SU"
    assert result.rule.dst_end
    assert isinstance(result.rule.dst_end, tz_rule.RuleDate)
    assert result.rule.dst_end.rrule_str == "FREQ=YEARLY;BYMONTH=11;BYDAY=1SU"


@pytest.mark.parametrize("key", zoneinfo.available_timezones())
def test_all_zoneinfo(key: str) -> None:
    """Verify that all available timezones in the system have valid tzdata."""
    if key.startswith("System") or key == "localtime" or key in IGNORED_TIMEZONES:
        return

    result = timezoneinfo.read(key)
    assert result.rule

    if result.rule.dst_start:
        assert result.rule.dst_end
        assert isinstance(result.rule.dst_start, tz_rule.RuleDate)
        assert isinstance(result.rule.dst_end, tz_rule.RuleDate)
        # Verify a rule can be constructed
        assert next(iter(result.rule.dst_start.as_rrule()))
        assert next(iter(result.rule.dst_end.as_rrule()))
    else:
        # Fixed offset
        assert result.rule
        assert result.rule.std
        assert result.rule.std.name
        assert not result.rule.dst

    # Verify there is a paresable tz rule
    timezoneinfo.TzInfo.from_timezoneinfo(result)



"""Tests for timeline related calendar eents."""

from __future__ import annotations

import datetime
import zoneinfo

import pytest
from ical.exceptions import CalendarParseError

from ical.calendar import Calendar
from ical.component import ComponentModel
from ical.exceptions import RecurrenceError
from ical.event import Event
from ical.parsing.property import ParsedProperty, ParsedPropertyParameter
from ical.parsing.component import parse_content
from ical.timeline import Timeline
from ical.todo import Todo
from ical.types.recur import Frequency, Recur, RecurrenceId, Weekday, WeekdayValue
from ical.recurrence import Recurrences


def test_from_contentlines() -> None:
    """Test parsing a recurrence rule from a string."""
    recurrences = Recurrences.from_basic_contentlines(
        [
            "DTSTART;TZID=America/New_York:20220802T060000",
            "RRULE:FREQ=DAILY;COUNT=3",
        ]
    )
    assert recurrences.rrule == [
        Recur(
            freq=Frequency.DAILY,
            count=3,
        )
    ]
    assert recurrences.dtstart == datetime.datetime(
        2022, 8, 2, 6, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/New_York")
    )


def test_from_contentlines_rdate() -> None:
    """Test parsing a recurrence rule with RDATE from a string."""

    lines = [
        "RRULE:FREQ=DAILY;COUNT=3",
        "RDATE;VALUE=DATE:20220803,20220804",
        "IGNORED:20250806",
    ]

    # parse using full ical parser
    content = [
        "BEGIN:RECURRENCE",
        *lines,
        "END:RECURRENCE",
    ]
    component = parse_content("\n".join(content))
    assert component
    orig_recurrences = Recurrences.parse_obj(component[0].as_dict())
    recurrences = Recurrences.from_basic_contentlines(lines)
    assert recurrences.rrule == [
        Recur(
            freq=Frequency.DAILY,
            count=3,
        )
    ]
    assert recurrences.rdate == [
        datetime.date(2022, 8, 3),
        datetime.date(2022, 8, 4),
    ]


@pytest.mark.parametrize("property", ["RDATE", "EXDATE"])
@pytest.mark.parametrize(
    ("date_value", "expected"),
    [
        ("{property}:20220803T060000", [datetime.datetime(2022, 8, 3, 6, 0, 0)]),
        (
            "{property}:20220803T060000,20220804T060000",
            [
                datetime.datetime(2022, 8, 3, 6, 0, 0),
                datetime.datetime(2022, 8, 4, 6, 0, 0),
            ],
        ),
        ("{property}:20220803", [datetime.date(2022, 8, 3)]),
        (
            "{property}:20220803,20220804",
            [datetime.date(2022, 8, 3), datetime.date(2022, 8, 4)],
        ),
        (
            "{property};VALUE=DATE:20220803,20220804",
            [datetime.date(2022, 8, 3), datetime.date(2022, 8, 4)],
        ),
        (
            "{property};VALUE=DATE-TIME:20220803T060000,20220804T060000",
            [
                datetime.datetime(2022, 8, 3, 6, 0, 0),
                datetime.datetime(2022, 8, 4, 6, 0, 0),
            ],
        ),
        (
            "{property}:20220803T060000Z,20220804T060000Z",
            [
                datetime.datetime(2022, 8, 3, 6, 0, 0, tzinfo=datetime.UTC),
                datetime.datetime(2022, 8, 4, 6, 0, 0, tzinfo=datetime.UTC),
            ],
        ),
        (
            "{property};TZID=America/New_York:19980119T020000",
            [
                datetime.datetime(
                    1998, 1, 19, 2, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/New_York")
                )
            ],
        ),
    ],
)
def test_from_contentlines_date_values(
    property: str, date_value: str, expected: list[datetime.datetime | datetime.date]
) -> None:
    """Test parsing a recurrence rule with RDATE from a string."""
    lines = [
        "RRULE:FREQ=DAILY;COUNT=3",
        date_value.format(property=property),
    ]

    # Parse using full ical parser with a fake component
    content = [
        "BEGIN:RECURRENCE",
        *lines,
        "END:RECURRENCE",
    ]
    # assert content == 'a'
    component = parse_content("\n".join(content))
    assert component
    orig_recurrences = Recurrences.parse_obj(component[0].as_dict())

    # Parse using optimized parser
    recurrences = Recurrences.from_basic_contentlines(lines)

    # Compare both approaches
    assert orig_recurrences == recurrences

    # Additionally assert expected values from test parameters
    assert recurrences.rrule == [
        Recur(
            freq=Frequency.DAILY,
            count=3,
        )
    ]
    attr = property.lower()
    assert getattr(recurrences, attr) == expected


@pytest.mark.parametrize(
    "contentlines",
    [
        ["RRULE;COUNT=3"],
        ["RRULE:COUNT=3;FREQ=invalid"],
        ["EXDATE"],
        ["RDATE"],
        ["RRULE;COUNT=3", "EXDATE"],
        ["EXDATE", "RDATE"],
        ["EXDATE:20220803T060000", "RDATE:"],
    ],
)
def test_from_invalid_contentlines(contentlines: list[str]) -> None:
    """Test parsing content lines that are not valid."""
    with pytest.raises(CalendarParseError):
        Recurrences.from_basic_contentlines(contentlines)


def test_as_rrule() -> None:
    """Test parsing a recurrence rule from a string."""
    recurrences = Recurrences.from_basic_contentlines(
        [
            "DTSTART:20220802T060000Z",
            "RRULE:FREQ=DAILY;COUNT=3",
            "EXDATE:20220803T060000Z",
        ]
    )
    assert list(recurrences.as_rrule()) == [
        datetime.datetime(2022, 8, 2, 6, 0, 0, tzinfo=datetime.UTC),
        datetime.datetime(2022, 8, 4, 6, 0, 0, tzinfo=datetime.UTC),
    ]


def test_as_rrule_with_rdate() -> None:
    """Test parsing a recurrence rule from a string."""
    recurrences = Recurrences.from_basic_contentlines(
        [
            "DTSTART:20220801",
            "RDATE:20220803",
            "RDATE:20220804",
            "RDATE:20220805",
        ]
    )
    assert list(recurrences.as_rrule()) == [
        datetime.date(2022, 8, 3),
        datetime.date(2022, 8, 4),
        datetime.date(2022, 8, 5),
    ]


def test_as_rrule_with_date() -> None:
    """Test parsing a recurrence rule from a string."""
    recurrences = Recurrences.from_basic_contentlines(
        [
            "RRULE:FREQ=DAILY;COUNT=3",
            "EXDATE:20220803T060000Z",
        ]
    )
    assert list(recurrences.as_rrule(datetime.datetime(2022, 8, 2, 6, 0, 0, tzinfo=datetime.UTC))) == [
        datetime.datetime(2022, 8, 2, 6, 0, 0, tzinfo=datetime.UTC),
        datetime.datetime(2022, 8, 4, 6, 0, 0, tzinfo=datetime.UTC),
    ]


def test_as_rrule_without_date() -> None:
    """Test parsing a recurrence rule from a string."""
    recurrences = Recurrences.from_basic_contentlines(
        [
            "RRULE:FREQ=DAILY;COUNT=3",
            "EXDATE:20220803T060000Z",
        ]
    )
    with pytest.raises(ValueError, match="dtstart is required"):
        list(recurrences.as_rrule())


def test_rrule_failure() -> None:
    """Test parsing a recurrence rule from a string."""
    recurrences = Recurrences.from_basic_contentlines(
        [
            "DTSTART:20220802T060000Z",
            "RRULE:FREQ=DAILY;COUNT=3",
            "EXDATE:20220803T060000",
        ]
    )
    with pytest.raises(RecurrenceError, match="can't compare offset-naive"):
        list(recurrences.as_rrule())


def test_ics() -> None:
    """Test parsing a recurrence rule from a string."""
    recurrences = Recurrences.from_basic_contentlines(
        [
            "DTSTART:20220802T060000Z",
            "RRULE:FREQ=DAILY;COUNT=3",
            "EXDATE:20220803T060000Z",
        ]
    )
    assert recurrences.ics() == [
        "DTSTART:20220802T060000Z",
        "RRULE:FREQ=DAILY;COUNT=3",
        "EXDATE:20220803T060000Z",
    ]



def test_mismatch_date_and_datetime_types() -> None:
    """Test parsing a recurrence rule from a string."""
    recurrences = Recurrences.from_basic_contentlines(
        [
            "DTSTART:20220801T060000Z",
            "RDATE:20220803",
            "RDATE:20220804T060000Z",
            "RDATE:20220805",
        ]
    )
    with pytest.raises(RecurrenceError):
        list(recurrences.as_rrule())



"""Tests for RELATED-TO data types."""


import pytest
from ical.exceptions import CalendarParseError

from ical.component import ComponentModel
from ical.parsing.component import ParsedComponent
from ical.parsing.property import ParsedProperty, ParsedPropertyParameter
from ical.types import RelatedTo, RelationshipType
from ical.types.data_types import DATA_TYPE


class FakeModel(ComponentModel):
    """Model under test."""

    example: RelatedTo

    class Config:
        """Pydantic model configuration."""

        json_encoders = DATA_TYPE.encode_property_json


def test_default_reltype() -> None:
    """Test for no explicit reltype specified."""
    model = FakeModel.parse_obj(
        {
            "example": [
                ParsedProperty(
                    name="example",
                    value="example-uid@example.com",
                    params=[ParsedPropertyParameter(name="RELTYPE", values=["PARENT"])],
                )
            ]
        },
    )
    assert model.example
    assert model.example.uid == "example-uid@example.com"
    assert model.example.reltype == "PARENT"


@pytest.mark.parametrize(
    "reltype",
    [
        ("PARENT"),
        ("CHILD"),
        ("SIBBLING"),
    ],
)
def test_reltype(reltype: str) -> None:
    """Test for no explicit reltype specified."""

    model = FakeModel.parse_obj(
        {
            "example": [
                ParsedProperty(
                    name="example",
                    value="example-uid@example.com",
                    params=[ParsedPropertyParameter(name="reltype", values=[reltype])],
                )
            ]
        },
    )
    assert model.example
    assert model.example.uid == "example-uid@example.com"
    assert model.example.reltype == reltype


def test_invalid_reltype() -> None:
    with pytest.raises(CalendarParseError):
        FakeModel.parse_obj(
            {
                "example": [
                    ParsedProperty(
                        name="example",
                        value="example-uid@example.com",
                        params=[
                            ParsedPropertyParameter(
                                name="reltype", values=["invalid-reltype"]
                            )
                        ],
                    )
                ]
            },
        )


def test_too_many_reltype_values() -> None:
    with pytest.raises(CalendarParseError):
        FakeModel.parse_obj(
            {
                "example": [
                    ParsedProperty(
                        name="example",
                        value="example-uid@example.com",
                        params=[
                            ParsedPropertyParameter(
                                name="reltype", values=["PARENT", "SIBBLING"]
                            )
                        ],
                    )
                ]
            },
        )


def test_encode_default_reltype() -> None:
    """Test encoded period."""

    model = FakeModel(example=RelatedTo(uid="example-uid@example.com"))
    assert model.__encode_component_root__() == ParsedComponent(
        name="FakeModel",
        properties=[
            ParsedProperty(
                name="example",
                value="example-uid@example.com",
                params=[ParsedPropertyParameter(name="RELTYPE", values=["PARENT"])],
            ),
        ],
    )


def test_encode_reltype() -> None:
    """Test encoded period."""

    model = FakeModel(
        example=RelatedTo(uid="example-uid@example.com", reltype=RelationshipType.CHILD)
    )
    assert model.__encode_component_root__() == ParsedComponent(
        name="FakeModel",
        properties=[
            ParsedProperty(
                name="example",
                value="example-uid@example.com",
                params=[ParsedPropertyParameter(name="RELTYPE", values=["CHILD"])],
            ),
        ],
    )



"""A Timeline is a set of events on a calendar.

A timeline can be used to iterate over all events, including expanded
recurring events. A timeline also supports methods to scan ranges of events
like returning all events happening today or after a specific date.
"""

from __future__ import annotations

import datetime
from collections.abc import Iterable, Iterator
from typing import TypeVar, Generic, Protocol

from .event import Event
from .iter import (
    SortableItemTimeline,
    SpanOrderedItem,
)
from .recur_adapter import merge_and_expand_items, ItemType

__all__ = ["Timeline", "generic_timeline"]

Timeline = SortableItemTimeline[Event]


def calendar_timeline(events: list[Event], tzinfo: datetime.tzinfo) -> Timeline:
    """Create a timeline for events on a calendar, including recurrence."""
    return Timeline(merge_and_expand_items(events, tzinfo))


def generic_timeline(
    items: list[ItemType], tzinfo: datetime.tzinfo
) -> SortableItemTimeline[ItemType]:
    """Return a timeline view of events on the calendar.

    All events are returned as if the attendee is viewing from the
    specified timezone. For example, this affects the order that All Day
    events are returned.
    """
    return SortableItemTimeline(
        merge_and_expand_items(
            items,
            tzinfo,
        )
    )



"""Tests for the tzif library."""

import datetime
from typing import cast

import pytest

from ical.tzif import tz_rule

TEST_DATETIME = datetime.datetime(2022, 1, 1)


def expand_rule(test_rule: tz_rule.RuleDate) -> datetime.datetime:
    """Test method to expand a rule to a single value."""
    return cast(datetime.datetime, next(iter(test_rule.as_rrule(TEST_DATETIME))))


def test_standard() -> None:
    """Test standard time with no daylight savings time."""
    rule = tz_rule.parse_tz_rule("EST5")
    assert rule.std.name == "EST"
    assert rule.std.offset == datetime.timedelta(hours=-5)
    assert rule.dst is None
    assert rule.dst_start is None
    assert rule.dst_end is None


def test_standard_plus_offset() -> None:
    """Test standard time with an offset with an explicit plus."""
    rule = tz_rule.parse_tz_rule("EST+5")
    assert rule.std.name == "EST"
    assert rule.std.offset == datetime.timedelta(hours=-5)
    assert rule.dst is None
    assert rule.dst_start is None
    assert rule.dst_end is None


def test_hours_minutes_offset() -> None:
    """Test standard time offset with hours and minutes."""
    rule = tz_rule.parse_tz_rule("EX05:30")
    assert rule.std.name == "EX"
    assert rule.std.offset == datetime.timedelta(hours=-5, minutes=-30)
    assert rule.dst is None
    assert rule.dst_start is None
    assert rule.dst_end is None


def test_hours_minutes_seconds_offset() -> None:
    """Test standard time offset with hours, minutes, and seconds."""
    rule = tz_rule.parse_tz_rule("EX05:30:20")
    assert rule.std.name == "EX"
    assert rule.std.offset == datetime.timedelta(hours=-5, minutes=-30, seconds=-20)
    assert rule.dst is None
    assert rule.dst_start is None
    assert rule.dst_end is None


def test_standard_minus_offset() -> None:
    """Test standard time offset with a negative offset."""
    rule = tz_rule.parse_tz_rule("JST-9")
    assert rule.std.name == "JST"
    assert rule.std.offset == datetime.timedelta(hours=9)
    assert rule.dst is None
    assert rule.dst_start is None
    assert rule.dst_end is None


def test_dst_implicit_offset() -> None:
    """Test daylight savings time with an implicit offset."""
    rule = tz_rule.parse_tz_rule("EST5EDT")
    assert rule.std.name == "EST"
    assert rule.std.offset == datetime.timedelta(hours=-5)
    assert rule.dst
    assert rule.dst.name == "EDT"
    assert rule.dst.offset == datetime.timedelta(hours=-4)
    assert rule.dst_start is None
    assert rule.dst_end is None


def test_standard_dst_implied_offset() -> None:
    """Test standard time with an offset with an explicit plus."""
    rule = tz_rule.parse_tz_rule("PST8PDT")
    assert rule.std.name == "PST"
    assert rule.std.offset == datetime.timedelta(hours=-8)
    assert rule.dst
    assert rule.dst.name == "PDT"
    assert rule.dst.offset == datetime.timedelta(hours=-7)
    assert rule.dst_start is None
    assert rule.dst_end is None


def test_dst_explicit_offset() -> None:
    """Test standard time with no daylight savings time."""
    rule = tz_rule.parse_tz_rule("EST5EDT4")
    assert rule.std.name == "EST"
    assert rule.std.offset == datetime.timedelta(hours=-5)
    assert rule.dst
    assert rule.dst.name == "EDT"
    assert rule.dst.offset == datetime.timedelta(hours=-4)
    assert rule.dst_start is None
    assert rule.dst_end is None


def test_dst_rules() -> None:
    """Test daylight savings start/end value."""
    rule = tz_rule.parse_tz_rule("EST+5EDT,M3.2.0/2,M11.1.0/2")
    assert rule.std.name == "EST"
    assert rule.std.offset == datetime.timedelta(hours=-5)
    assert rule.dst
    assert rule.dst.name == "EDT"
    assert rule.dst.offset == datetime.timedelta(hours=-4)
    assert rule.dst_start
    assert isinstance(rule.dst_start, tz_rule.RuleDate)
    assert rule.dst_start.month == 3
    assert rule.dst_start.week_of_month == 2
    assert rule.dst_start.day_of_week == 0
    assert rule.dst_start.time == datetime.timedelta(hours=2)
    assert rule.dst_end
    assert isinstance(rule.dst_end, tz_rule.RuleDate)
    assert rule.dst_end.month == 11
    assert rule.dst_end.week_of_month == 1
    assert rule.dst_end.day_of_week == 0
    assert rule.dst_end.time == datetime.timedelta(hours=2)

    assert next(
        iter(rule.dst_start.as_rrule(datetime.datetime(2022, 1, 1)))
    ) == datetime.datetime(2022, 3, 13, 2, 0, 0)
    assert next(
        iter(rule.dst_end.as_rrule(datetime.datetime(2022, 1, 1)))
    ) == datetime.datetime(2022, 11, 6, 2, 0, 0)

    assert rule.dst_start.rrule_str == "FREQ=YEARLY;BYMONTH=3;BYDAY=2SU"
    assert rule.dst_end.rrule_str == "FREQ=YEARLY;BYMONTH=11;BYDAY=1SU"


def test_dst_implement_time_rules() -> None:
    """Test daylight savings values rules with no explicit time."""
    rule = tz_rule.parse_tz_rule("EST+5EDT,M3.2.0,M11.1.0")
    assert rule.std.name == "EST"
    assert rule.std.offset == datetime.timedelta(hours=-5)
    assert rule.dst
    assert rule.dst.name == "EDT"
    assert rule.dst.offset == datetime.timedelta(hours=-4)
    assert rule.dst_start
    assert isinstance(rule.dst_start, tz_rule.RuleDate)
    assert rule.dst_start.month == 3
    assert rule.dst_start.week_of_month == 2
    assert rule.dst_start.day_of_week == 0
    assert rule.dst_start.time == datetime.timedelta(hours=2)
    assert rule.dst_end
    assert isinstance(rule.dst_end, tz_rule.RuleDate)
    assert rule.dst_end.month == 11
    assert rule.dst_end.week_of_month == 1
    assert rule.dst_end.day_of_week == 0
    assert rule.dst_end.time == datetime.timedelta(hours=2)


@pytest.mark.parametrize(
    "tz_string",
    [
        "",
        "1234",
        "EST+5EDT,M3.2.0/2",
        "EST+5EDT,M3.2.0/2,M11.1.0/2,M3",
        "EST+5EDT,3.2.0/2,M11.1.0/2",
        "EST+5EDT,M3.2/2,M11.1.0/2",
        "EST+5EDT,M3.2.0.4/2,M11.1.0/2",
    ],
)
def test_invalid(tz_string: str) -> None:
    """Test an invalid rule occurrence"""
    with pytest.raises(ValueError, match="Unable to parse TZ string"):
        tz_rule.parse_tz_rule(tz_string)


def test_tz_offset() -> None:
    """Test standard time offset with hours and minutes."""
    rule = tz_rule.parse_tz_rule("<-03>3<-02>,M3.5.0/-2,M10.5.0/-1")
    assert rule.std.name == "<-03>"
    assert rule.std.offset == datetime.timedelta(hours=-3)
    assert rule.dst
    assert rule.dst.name == "<-02>"
    assert rule.dst.offset == datetime.timedelta(hours=-2)
    assert rule.dst_start
    assert isinstance(rule.dst_start, tz_rule.RuleDate)
    assert rule.dst_start.month == 3
    assert rule.dst_start.week_of_month == 5
    assert rule.dst_start.day_of_week == 0
    assert rule.dst_start.time == datetime.timedelta(hours=-2)
    assert rule.dst_end
    assert isinstance(rule.dst_end, tz_rule.RuleDate)
    assert rule.dst_end.month == 10
    assert rule.dst_end.week_of_month == 5
    assert rule.dst_end.day_of_week == 0
    assert rule.dst_end.time == datetime.timedelta(hours=-1)


def test_iran_rule_offset() -> None:
    """Test a more complex timezone rule."""
    rule = tz_rule.parse_tz_rule("<+0330>-3:30<+0430>,J79/24,J263/24")
    assert rule.std.name == "<+0330>"
    assert rule.std.offset == datetime.timedelta(hours=3, minutes=30)
    assert rule.dst
    assert rule.dst.name == "<+0430>"
    assert rule.dst.offset == datetime.timedelta(hours=4, minutes=30)
    assert rule.dst_start
    assert isinstance(rule.dst_start, tz_rule.RuleDay)
    assert rule.dst_start.day_of_year == 79
    assert rule.dst_start.time == datetime.timedelta(hours=24)
    assert rule.dst_end
    assert isinstance(rule.dst_end, tz_rule.RuleDay)
    assert rule.dst_end.day_of_year == 263
    assert rule.dst_end.time == datetime.timedelta(hours=24)


def test_invalid_time() -> None:
    """Test validation of fields with an invalid time value."""
    with pytest.raises(ValueError, match="time was not parse tree dict"):
        tz_rule.RuleDate.parse_obj(
            {
                "month": 3,
                "week_of_month": 1,
                "day_of_week": 0,
                "time": 0.12345,
            }
        )



"""Tests for timeline related calendar events."""

from collections.abc import Generator
import itertools
import json
import textwrap
import pathlib

import pytest
from syrupy import SnapshotAssertion

from ical.exceptions import CalendarParseError
from ical.calendar_stream import CalendarStream, IcsCalendarStream
from ical.store import TodoStore

MAX_ITERATIONS = 30
TESTDATA_PATH = pathlib.Path("tests/testdata/")
TESTDATA_FILES = list(TESTDATA_PATH.glob("*.ics"))
TESTDATA_IDS = [ x.stem for x in TESTDATA_FILES ]


def test_empty_ics(mock_prodid: Generator[None, None, None]) -> None:
    """Test serialization of an empty ics file."""
    calendar = IcsCalendarStream.calendar_from_ics("")
    ics = IcsCalendarStream.calendar_to_ics(calendar)
    assert (
        ics
        == textwrap.dedent("""\
            BEGIN:VCALENDAR
            PRODID:-//example//1.2.3
            VERSION:2.0
            END:VCALENDAR"""
    ))

    calendar.prodid = "-//example//1.2.4"
    ics = IcsCalendarStream.calendar_to_ics(calendar)
    assert (
        ics
        == textwrap.dedent("""\
            BEGIN:VCALENDAR
            PRODID:-//example//1.2.4
            VERSION:2.0
            END:VCALENDAR""")
    )


@pytest.mark.parametrize("filename", TESTDATA_FILES, ids=TESTDATA_IDS)
def test_parse(filename: pathlib.Path, snapshot: SnapshotAssertion, json_encoder: json.JSONEncoder) -> None:
    """Fixture to read golden file and compare to golden output."""
    cal = CalendarStream.from_ics(filename.read_text())
    data = json.loads(
        cal.json(exclude_unset=True, exclude_none=True, encoder=json_encoder.default)
    )
    assert snapshot == data

    # Re-parse the data object to verify we get the original data values
    # back. This effectively confirms that all fields can be parsed from the
    # python native format in addition to rfc5545.
    cal_reparsed = CalendarStream.parse_obj(data)
    data_reparsed = json.loads(
        cal_reparsed.json(
            exclude_unset=True, exclude_none=True, encoder=json_encoder.default
        )
    )
    assert data_reparsed == data


@pytest.mark.parametrize("filename", TESTDATA_FILES, ids=TESTDATA_IDS)
def test_serialize(filename: pathlib.Path, snapshot: SnapshotAssertion) -> None:
    """Fixture to read golden file and compare to golden output."""
    with filename.open() as f:
        cal = IcsCalendarStream.from_ics(f.read())
    assert cal.ics() == snapshot


@pytest.mark.parametrize("filename", TESTDATA_FILES, ids=TESTDATA_IDS)
def test_timeline_iteration(filename: pathlib.Path) -> None:
    """Fixture to ensure all calendar events are valid and support iteration."""
    with filename.open() as f:
        cal = IcsCalendarStream.from_ics(f.read())
    for calendar in cal.calendars:
        # Iterate over the timeline to ensure events are valid. There is a max
        # to handle recurring events that may repeat forever.
        for event in itertools.islice(calendar.timeline, MAX_ITERATIONS):
            assert event is not None


@pytest.mark.parametrize("filename", TESTDATA_FILES, ids=TESTDATA_IDS)
def test_todo_list_iteration(filename: pathlib.Path) -> None:
    """Fixture to read golden file and compare to golden output."""
    cal = CalendarStream.from_ics(filename.read_text())
    if not cal.calendars:
        return
    calendar = cal.calendars[0]
    tl = TodoStore(calendar).todo_list()
    for todo in itertools.islice(tl, MAX_ITERATIONS):
        assert todo is not None


def test_invalid_ics() -> None:
    """Test parsing failures for ics content."""
    with pytest.raises(CalendarParseError, match="Failed to parse calendar stream"):
        IcsCalendarStream.calendar_from_ics("invalid")


def test_component_failure() -> None:
    with pytest.raises(CalendarParseError, match="Failed to parse component"):
        IcsCalendarStream.calendar_from_ics(
            textwrap.dedent("""\
                BEGIN:VCALENDAR
                PRODID:-//example//1.2.3
                VERSION:2.0
                BEGIN:VEVENT
                DTSTART:20220724T120000
                DTEND:20220724
                END:VEVENT
                END:VCALENDAR
            """))
        

def test_multiple_calendars() -> None:
    with pytest.raises(CalendarParseError, match="more than one calendar"):
        IcsCalendarStream.calendar_from_ics(
            textwrap.dedent("""\
                BEGIN:VCALENDAR
                PRODID:-//example//1.2.3
                VERSION:2.0
                END:VCALENDAR
                BEGIN:VCALENDAR
                PRODID:-//example//1.2.3
                VERSION:2.0
                END:VCALENDAR
            """))



"""Tests for handling rfc5545 properties and parameters.

This reuses the test data made for testing full components, but also exercises
the same lower level components.
"""

from dataclasses import asdict
import pathlib

import pytest
from syrupy import SnapshotAssertion

from ical.parsing.property import (
    ParsedProperty,
    ParsedPropertyParameter,
    parse_basic_ics_properties,
)
from ical.parsing.component import unfolded_lines


TESTDATA_PATH = pathlib.Path("tests/parsing/testdata/")
TESTDATA_FILES = list(TESTDATA_PATH.glob("*.ics"))
TESTDATA_IDS = [x.stem for x in TESTDATA_FILES]


@pytest.mark.parametrize("filename", TESTDATA_FILES, ids=TESTDATA_IDS)
def test_from_ics(filename: str, snapshot: SnapshotAssertion) -> None:
    """Fixture to read golden file and compare to golden output."""
    properties = list(parse_basic_ics_properties(unfolded_lines(filename.read_text())))
    assert properties == snapshot

@pytest.mark.parametrize(
    "ics",
    [
        "PROP-VALUE",
        "PROP;:VALUE",
        "PROP;PARAM:VALUE",
        ";VALUE",
        ";:VALUE",
    ]
)
def test_invalid_format(ics: str) -> None:
    """Test parsing invalid property format."""
    with pytest.raises(ValueError):
        r = list(parse_basic_ics_properties([ics]))
        assert r == 'a'



"""Component specific iterable functions.

This module contains functions that are helpful for iterating over components
in a calendar. This includes expanding recurring components or other functions
for managing components from a list (e.g. grouping by uid).
"""

import datetime
from collections.abc import Iterable
from typing import Generic, TypeVar, cast

from .iter import (
    MergedIterable,
    RecurIterable,
    SortableItemValue,
    SpanOrderedItem,
    LazySortableItem,
    SortableItem,
)
from .types.recur import RecurrenceId
from .event import Event
from .todo import Todo
from .journal import Journal
from .freebusy import FreeBusy
from .timespan import Timespan


ItemType = TypeVar("ItemType", bound="Event | Todo | Journal")
_DateOrDatetime = datetime.datetime | datetime.date


class RecurAdapter(Generic[ItemType]):
    """An adapter that expands an Event instance for a recurrence rule.

    This adapter is given an event, then invoked with a specific date/time instance
    that the event occurs on due to a recurrence rule. The event is copied with
    necessary updated fields to act as a flattened instance of the event.
    """

    def __init__(
        self,
        item: ItemType,
        tzinfo: datetime.tzinfo | None = None,
    ) -> None:
        """Initialize the RecurAdapter."""
        self._item = item
        self._duration = item.computed_duration
        self._tzinfo = tzinfo

    def get(
        self, dtstart: datetime.datetime | datetime.date
    ) -> SortableItem[Timespan, ItemType]:
        """Return a lazy sortable item."""

        recur_id_dt = dtstart
        dtend = dtstart + self._duration if self._duration else dtstart
        # Make recurrence_id floating time to avoid dealing with serializing
        # TZID. This value will still be unique within the series and is in
        # the context of dtstart which may have a timezone.
        if isinstance(recur_id_dt, datetime.datetime) and recur_id_dt.tzinfo:
            recur_id_dt = recur_id_dt.replace(tzinfo=None)
        recurrence_id = RecurrenceId.__parse_property_value__(recur_id_dt)

        def build() -> ItemType:
            updates = {
                "dtstart": dtstart,
                "recurrence_id": recurrence_id,
            }
            if isinstance(self._item, Event) and self._item.dtend and dtend:
                updates["dtend"] = dtend
            if isinstance(self._item, Todo) and self._item.due and dtend:
                updates["due"] = dtend
            return cast(ItemType, self._item.copy(update=updates))

        ts = Timespan.of(dtstart, dtend, self._tzinfo)
        return LazySortableItem(ts, build)


def items_by_uid(items: list[ItemType]) -> dict[str, list[ItemType]]:
    items_by_uid: dict[str, list[ItemType]] = {}
    for item in items:
        if item.uid is None:
            raise ValueError("Todo must have a UID")
        if (values := items_by_uid.get(item.uid)) is None:
            values = []
            items_by_uid[item.uid] = values
        values.append(item)
    return items_by_uid


def merge_and_expand_items(
    items: list[ItemType], tzinfo: datetime.tzinfo
) -> Iterable[SpanOrderedItem[ItemType]]:
    """Merge and expand items that are recurring."""
    iters: list[Iterable[SpanOrderedItem[ItemType]]] = []
    for item in items:
        if not (recur := item.as_rrule()):
            iters.append(
                [
                    SortableItemValue(
                        item.timespan_of(tzinfo),
                        item,
                    )
                ]
            )
            continue
        iters.append(RecurIterable(RecurAdapter(item, tzinfo=tzinfo).get, recur))
    return MergedIterable(iters)



"""A List is a set of objects on a calendar.

A List is used to iterate over all objects, including expanded recurring
objects. A List is similar to a Timeline, except it does not repeat recurring
objects on the list and they are only shown once. A list does not repeat
forever.
"""

import datetime
from collections.abc import Generator
import logging

from .todo import Todo
from .recur_adapter import items_by_uid, merge_and_expand_items
from .util import local_timezone

# Not part of the public API
__all__: list[str] = []

_LOGGER = logging.getLogger(__name__)


def _pick_todo(todos: list[Todo], dtstart: datetime.datetime) -> Todo:
    """Pick a todo to return in a list from a list of recurring todos.

    The items passed in must all be for the same original todo (either a
    single todo or instance of a recurring todo including any edits). An
    edited instance of a recurring todo has a recurrence-id that is
    different from the original todo. This function will return the
    next todo that is incomplete and has the latest due date.
    """
    # For a recurring todo, the dtstart is after the last due date. Therefore
    # we can stort items by dtstart and pick the last one that hasn't happened
    root_iter = merge_and_expand_items(todos, dtstart.tzinfo or local_timezone())

    it = iter(root_iter)
    last = next(it)
    while cur := next(it, None):
        if cur.item.start_datetime is None or cur.item.start_datetime > dtstart:
            break
        last = cur
    return last.item


def todo_list_view(
    todos: list[Todo],
    dtstart: datetime.datetime | None = None,
) -> Generator[Todo, None, None]:
    """Create a list view for todos on a calendar, including recurrence.

    The dtstart value is used to determine the current time for the list and
    for deciding which instance of a recurring todo to return.
    """
    if dtstart is None:
        dtstart = datetime.datetime.now(tz=local_timezone())
    todos_by_uid = items_by_uid(todos)
    for todos in todos_by_uid.values():
        yield _pick_todo(todos, dtstart)



"""Tests for PRIORITY types."""

import pytest
from ical.exceptions import CalendarParseError

from ical.component import ComponentModel
from ical.parsing.property import ParsedProperty
from ical.types import Priority


class FakeModel(ComponentModel):
    """Model under test."""

    pri: Priority


def test_priority() -> None:
    """Test for priority fields."""

    model = FakeModel.parse_obj({"pri": [ParsedProperty(name="dt", value="1")]})
    assert model.pri == 1

    model = FakeModel.parse_obj({"pri": [ParsedProperty(name="dt", value="9")]})
    assert model.pri == 9

    with pytest.raises(CalendarParseError):
        FakeModel.parse_obj({"pri": [ParsedProperty(name="dt", value="-1")]})

    with pytest.raises(CalendarParseError):
        FakeModel.parse_obj({"pri": [ParsedProperty(name="dt", value="10")]})
